#ifndef RK_DNSPACK_BASISCOMPOSITION_H
#define RK_DNSPACK_BASISCOMPOSITION_H

#include <map>
#include <set>
#include <stdexcept>


/*
	BasisComposition may contains the empty m_composition.
	However, you cannot get the abnormal composition.
*/


namespace rk
{
	namespace dnspack
	{
		class BasisComposition
		{
		public:
			BasisComposition();
			BasisComposition(const std::map<double, int>& composition);
			BasisComposition(std::map<double, int>&& composition);
			virtual ~BasisComposition();

		public:
			BasisComposition(const BasisComposition&) = default;
			BasisComposition(BasisComposition&&) noexcept = default;
			BasisComposition& operator=(const BasisComposition&) = default;
			BasisComposition& operator=(BasisComposition&&) noexcept = default;

			bool operator==(const BasisComposition& other) const;
			bool operator!=(const BasisComposition& other) const;

		public:
			std::size_t getSphereKindNumber() const;
			std::size_t getSpheresNumber() const;
			std::set<double> getRadii() const;
			const std::map<double, int>& getComposition() const;

			void setComposition(const std::map<double, int>& composition);
			void setComposition(std::map<double, int>&& composition);
			void clear() noexcept;

			bool empty() const noexcept;

		private:
			std::map<double, int> m_composition; // radius and the number in an unit cell
		};
	}
}


inline bool rk::dnspack::BasisComposition::operator==(const BasisComposition& other) const
{
	return (m_composition == other.m_composition);
}

inline bool rk::dnspack::BasisComposition::operator!=(const BasisComposition& other) const
{
	return !(*this == other);
}

inline std::size_t rk::dnspack::BasisComposition::getSphereKindNumber() const
{
	if (m_composition.empty())
	{
		throw std::logic_error("m_composition is empty.");
	}

	return m_composition.size();
}

inline std::size_t rk::dnspack::BasisComposition::getSpheresNumber() const
{
	if (m_composition.empty())
	{
		throw std::logic_error("m_composition is empty.");
	}

	std::size_t spheresNumber = 0;

	for (const auto& number : m_composition)
	{
		spheresNumber += number.second;
	}

	return spheresNumber;
}

inline std::set<double> rk::dnspack::BasisComposition::getRadii() const
{
	if (m_composition.empty())
	{
		throw std::logic_error("m_composition is empty.");
	}

	std::set<double> radii;

	for (const auto& number : m_composition)
	{
		radii.emplace(number.first);
	}
	
	return std::move(radii);
}

inline const std::map<double, int>& rk::dnspack::BasisComposition::getComposition() const
{
	if (m_composition.empty())
	{
		throw std::logic_error("m_composition is empty.");
	}

	return m_composition;
}

inline void rk::dnspack::BasisComposition::setComposition(const std::map<double, int>& composition)
{
	for (const auto& number : composition)
	{
		if (number.first <= 0.0)
		{
			throw std::invalid_argument("composition has a minus radius.");
		}

		if (number.second <= 0)
		{
			throw std::invalid_argument("composition has a minus number of each sphere in an unit cell.");
		}
	}

	m_composition = composition;
}

inline void rk::dnspack::BasisComposition::setComposition(std::map<double, int>&& composition)
{
	for (const auto& number : composition)
	{
		if (number.first <= 0.0)
		{
			throw std::invalid_argument("composition has a minus radius.");
		}

		if (number.second <= 0)
		{
			throw std::invalid_argument("composition has a minus number of each sphere in an unit cell.");
		}
	}

	m_composition = std::move(composition);
}

inline void rk::dnspack::BasisComposition::clear() noexcept
{
	m_composition.clear();
}

inline bool rk::dnspack::BasisComposition::empty() const noexcept
{
	return (m_composition.empty());
}


#endif // !RK_DNSPACK_BASISCOMPOSITION_H
