#ifndef RK_DNSPACK_CALCULATIONANALYSIS_H
#define RK_DNSPACK_CALCULATIONANALYSIS_H

#include <string>
#include <mutex>
#include <map>


namespace rk
{
	namespace dnspack
	{
		class CalculationAnalysis
		{
		public:
			CalculationAnalysis();
			virtual ~CalculationAnalysis();

		public:
			CalculationAnalysis(const CalculationAnalysis&) = default;
			CalculationAnalysis(CalculationAnalysis&&) noexcept = default;
			CalculationAnalysis& operator=(const CalculationAnalysis&) = default;
			CalculationAnalysis& operator=(CalculationAnalysis&&) noexcept = default;

		public:
			std::map<double, int> getPackingFractionsErrorDistribution() const noexcept;
			std::map<double, int> getInitialPackingOptimizePackingFractionsDistribution() const noexcept;
			std::map<double, int> getPackingOptimizePackingFractionsDistribution() const noexcept;

			double getThresholdPackingFraction() const noexcept;
			int getLowPackingFractionRejectNumber() const noexcept;

			void appendPackingFractionsError(const double smallAnnealingPackingFraction, const double packingOptimizePackingFraction);
			void appendInitialPackingOptimizePackingFractionsDistribution(const double initialPackingOptimizePackingFraction);
			void appendPackingOptimizePackingFractionsDistribution(const double packingOptimizePackingFraction);

			void setThresholdPackingFraction();
			void appendLowPackingFractionRejectNumber() noexcept;

			int getAnnealingMinusVolumeRejectNumber() const noexcept;
			int getInitialPackingOptimizeMinusVolumeRejectNumber() const noexcept;
			int getPackingOptimizeMinusVolumeRejectNumber() const noexcept;
			int getPackingOptimizeZvalueRejectNumber() const noexcept;

			void appendAnnealingMinusVolumeRejectNumber() noexcept;
			void appendInitialPackingOptimizeMinusVolumeRejectNumber() noexcept;
			void appendPackingOptimizeMinusVolumeRejectNumber() noexcept;
			void appendPackingOptimizeZvalueRejectNumber() noexcept;

		public:
			static void setPackingFractionsErrorMesh(const double packingFractionsErrorMesh) noexcept;
			static void setInitialPackingOptimizePackingFractionsMesh(const double initialPackingOptimizePackingFractionsMesh) noexcept;
			static void setPackingOptimizePackingFractionsMesh(const double packingOptimizePackingFractionsMesh) noexcept;

			static void clear() noexcept;

		private:
			static double m_packingFractionsErrorMesh;
			static double m_initialPackingOptimizePackingFractionsMesh;
			static double m_packingOptimizePackingFractionsMesh;

			static std::map<double, int> m_packingFractionsErrorDistribution;
			static std::map<double, int> m_initialPackingOptimizePackingFractionsDistribution;
			static std::map<double, int> m_packingOptimizePackingFractionsDistribution;

			static double m_thresholdPackingFraction;
			static int m_lowPackingFractionRejectNumber;

			static int m_annealingMinusVolumeRejectNumber;
			static int m_initialPackingOptimizeMinusVolumeRejectNumber;
			static int m_packingOptimizeMinusVolumeRejectNumber;
			static int m_packingOptimizeZvalueRejectNumber;
			
			static std::mutex m_analysisMutex;
		};
	}
}


inline std::map<double, int> rk::dnspack::CalculationAnalysis::getPackingFractionsErrorDistribution() const noexcept
{
	return m_packingFractionsErrorDistribution;
}

inline std::map<double, int> rk::dnspack::CalculationAnalysis::getInitialPackingOptimizePackingFractionsDistribution() const noexcept
{
	return m_initialPackingOptimizePackingFractionsDistribution;
}

inline std::map<double, int> rk::dnspack::CalculationAnalysis::getPackingOptimizePackingFractionsDistribution() const noexcept
{
	return m_packingOptimizePackingFractionsDistribution;
}

inline double rk::dnspack::CalculationAnalysis::getThresholdPackingFraction() const noexcept
{
	return m_thresholdPackingFraction;
}

inline int rk::dnspack::CalculationAnalysis::getLowPackingFractionRejectNumber() const noexcept
{
	return m_lowPackingFractionRejectNumber;
}

inline void rk::dnspack::CalculationAnalysis::appendPackingFractionsError(const double smallAnnealingPackingFraction, const double packingOptimizePackingFraction)
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };

	double error = packingOptimizePackingFraction - smallAnnealingPackingFraction;
	
	double labelError = -20.0;
	{
		int label = static_cast<int>(error / m_packingFractionsErrorMesh);

		if (error < 0)
		{
			labelError = (static_cast<double>(label) - 0.5) * m_packingFractionsErrorMesh;
		}

		else
		{
			labelError = (static_cast<double>(label) + 0.5) * m_packingFractionsErrorMesh;
		}
	}

	std::map<double, int>::iterator iter = m_packingFractionsErrorDistribution.find(labelError);
	{
		if (m_packingFractionsErrorDistribution.end() == iter)
		{
			m_packingFractionsErrorDistribution.emplace(labelError, 1);
		}

		else
		{
			iter->second += 1;
		}
	}
}

inline void rk::dnspack::CalculationAnalysis::appendInitialPackingOptimizePackingFractionsDistribution(const double smallAnnealingPackingFraction)
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };

	double tempPackingFraction = smallAnnealingPackingFraction - 0.70;

	double labelPackingFraction = 0.0;
	{
		int label = static_cast<int>(tempPackingFraction / m_initialPackingOptimizePackingFractionsMesh);

		if (tempPackingFraction < 0)
		{
			labelPackingFraction = (static_cast<double>(label) - 0.5) * m_initialPackingOptimizePackingFractionsMesh + 0.70;
		}

		else
		{
			labelPackingFraction = (static_cast<double>(label) + 0.5) * m_initialPackingOptimizePackingFractionsMesh + 0.70;
		}
	}

	std::map<double, int>::iterator iter = m_initialPackingOptimizePackingFractionsDistribution.find(labelPackingFraction);

	if (m_initialPackingOptimizePackingFractionsDistribution.end() == iter)
	{
		m_initialPackingOptimizePackingFractionsDistribution.emplace(labelPackingFraction, 1);
	}

	else
	{
		iter->second += 1;
	}
}

inline void rk::dnspack::CalculationAnalysis::appendPackingOptimizePackingFractionsDistribution(const double packingOptimizePackingFraction)
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };

	double tempPackingFraction = packingOptimizePackingFraction - 0.70;

	double labelPackingFraction = 0.0;
	{
		int label = static_cast<int>(tempPackingFraction / m_packingOptimizePackingFractionsMesh);

		if (tempPackingFraction < 0)
		{
			labelPackingFraction = (static_cast<double>(label) - 0.5) * m_packingOptimizePackingFractionsMesh + 0.70;
		}

		else
		{
			labelPackingFraction = (static_cast<double>(label) + 0.5) * m_packingOptimizePackingFractionsMesh + 0.70;
		}
	}

	std::map<double, int>::iterator iter = m_packingOptimizePackingFractionsDistribution.find(labelPackingFraction);

	if (m_packingOptimizePackingFractionsDistribution.end() == iter)
	{
		m_packingOptimizePackingFractionsDistribution.emplace(labelPackingFraction, 1);
	}

	else
	{
		iter->second += 1;
	}
}

inline void rk::dnspack::CalculationAnalysis::setThresholdPackingFraction()
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };

	if (m_packingOptimizePackingFractionsDistribution.empty())
	{
		throw std::logic_error("m_packingOptimizePackingFractionsDistribution is empty.");
	}

	int num = 0;
	double error = 0.0;

	int label = 1;

	for (const auto& distribution : m_packingFractionsErrorDistribution)
	{
		if (distribution.first < 0)
		{
			int errorLabel = static_cast<int>(distribution.first / m_packingFractionsErrorMesh);
			errorLabel *= -1;
			errorLabel += 1;

			num += distribution.second * errorLabel;
			error -= distribution.first * static_cast<double>(distribution.second) * errorLabel;
		}

		else
		{
			num += distribution.second * label;
			error += distribution.first * static_cast<double>(distribution.second) * label;

			++label;
		}
	}

	error /= static_cast<double>(num);

	std::map<double, int>::iterator iterPrevEnd = std::prev(m_packingOptimizePackingFractionsDistribution.end());
	double maxPackingFraction = iterPrevEnd->first - 0.5 * m_packingOptimizePackingFractionsMesh;

	m_thresholdPackingFraction = maxPackingFraction - error - 0.02;
}

inline void rk::dnspack::CalculationAnalysis::appendLowPackingFractionRejectNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };
	m_lowPackingFractionRejectNumber += 1;
}

inline int rk::dnspack::CalculationAnalysis::getAnnealingMinusVolumeRejectNumber() const noexcept
{
	return m_annealingMinusVolumeRejectNumber;
}

inline int rk::dnspack::CalculationAnalysis::getInitialPackingOptimizeMinusVolumeRejectNumber() const noexcept
{
	return m_initialPackingOptimizeMinusVolumeRejectNumber;
}

inline int rk::dnspack::CalculationAnalysis::getPackingOptimizeMinusVolumeRejectNumber() const noexcept
{
	return m_packingOptimizeMinusVolumeRejectNumber;
}

inline int rk::dnspack::CalculationAnalysis::getPackingOptimizeZvalueRejectNumber() const noexcept
{
	return m_packingOptimizeZvalueRejectNumber;
}

inline void rk::dnspack::CalculationAnalysis::appendAnnealingMinusVolumeRejectNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };
	m_annealingMinusVolumeRejectNumber += 1;
}

inline void rk::dnspack::CalculationAnalysis::appendInitialPackingOptimizeMinusVolumeRejectNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };
	m_initialPackingOptimizeMinusVolumeRejectNumber += 1;
}

inline void rk::dnspack::CalculationAnalysis::appendPackingOptimizeMinusVolumeRejectNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };
	m_packingOptimizeMinusVolumeRejectNumber += 1;
}

inline void rk::dnspack::CalculationAnalysis::appendPackingOptimizeZvalueRejectNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_analysisMutex };
	m_packingOptimizeZvalueRejectNumber += 1;
}

inline void rk::dnspack::CalculationAnalysis::setPackingFractionsErrorMesh(const double packingFractionsErrorMesh) noexcept
{
	m_packingFractionsErrorMesh = packingFractionsErrorMesh;
}

inline void rk::dnspack::CalculationAnalysis::setInitialPackingOptimizePackingFractionsMesh(const double initialPackingOptimizePackingFractionsMesh) noexcept
{
	m_initialPackingOptimizePackingFractionsMesh = initialPackingOptimizePackingFractionsMesh;
}

inline void rk::dnspack::CalculationAnalysis::setPackingOptimizePackingFractionsMesh(const double packingOptimizePackingFractionsMesh) noexcept
{
	m_packingOptimizePackingFractionsMesh = packingOptimizePackingFractionsMesh;
}

inline void rk::dnspack::CalculationAnalysis::clear() noexcept
{
	m_packingFractionsErrorDistribution.clear();
	m_initialPackingOptimizePackingFractionsDistribution.clear();
	m_packingOptimizePackingFractionsDistribution.clear();

	m_thresholdPackingFraction = 0.0;
	m_lowPackingFractionRejectNumber = 0;

	m_annealingMinusVolumeRejectNumber = 0;
	m_initialPackingOptimizeMinusVolumeRejectNumber = 0;
	m_packingOptimizeMinusVolumeRejectNumber = 0;
	m_packingOptimizeZvalueRejectNumber = 0;
}


#endif // !RK_DNSPACK_CALCULATIONANALYSIS_H
