#ifndef RK_DNSPACK_CARTESIANCOORDINATE_H
#define RK_DNSPACK_CARTESIANCOORDINATE_H

#include <stdexcept>
#include <cmath>

namespace rk
{
	namespace dnspack
	{
		class CartesianCoordinate
		{
		public:
			double x;
			double y;
			double z;

		public:
			CartesianCoordinate() noexcept;
			CartesianCoordinate(const double a, const double b, const double c) noexcept;
			virtual ~CartesianCoordinate();

		public:
			CartesianCoordinate(const CartesianCoordinate& other) noexcept;
			CartesianCoordinate(CartesianCoordinate&& other) noexcept;
			CartesianCoordinate& operator=(const CartesianCoordinate& other) noexcept;
			CartesianCoordinate& operator=(CartesianCoordinate&& other) noexcept;

			bool operator==(const CartesianCoordinate& other) const noexcept;
			bool operator!=(const CartesianCoordinate& other) const noexcept;

			CartesianCoordinate& operator+=(const CartesianCoordinate& other) noexcept;
			CartesianCoordinate& operator-=(const CartesianCoordinate& other) noexcept;
			CartesianCoordinate& scaling(const double num) noexcept;

			double operator[](const int direction) const;

		public:
			double getNorm() const noexcept;
			double getNormSquare() const noexcept;
		};
	}
}


inline rk::dnspack::CartesianCoordinate& rk::dnspack::CartesianCoordinate::operator=(const rk::dnspack::CartesianCoordinate& other) noexcept
{
	x = other.x;
	y = other.y;
	z = other.z;

	return *this;
}

inline rk::dnspack::CartesianCoordinate& rk::dnspack::CartesianCoordinate::operator=(rk::dnspack::CartesianCoordinate&& other) noexcept
{
	x = other.x;
	y = other.y;
	z = other.z;

	return *this;
}

inline bool rk::dnspack::CartesianCoordinate::operator==(const rk::dnspack::CartesianCoordinate& other) const noexcept
{
	return (x == other.x && y == other.y && z == other.z);
}

inline bool rk::dnspack::CartesianCoordinate::operator!=(const rk::dnspack::CartesianCoordinate& other) const noexcept
{
	return !(*this == other);
}

inline rk::dnspack::CartesianCoordinate& rk::dnspack::CartesianCoordinate::operator+=(const rk::dnspack::CartesianCoordinate& other) noexcept
{
	x += other.x;
	y += other.y;
	z += other.z;

	return *this;
}

inline rk::dnspack::CartesianCoordinate& rk::dnspack::CartesianCoordinate::operator-=(const rk::dnspack::CartesianCoordinate& other) noexcept
{
	x -= other.x;
	y -= other.y;
	z -= other.z;

	return *this;
}

inline rk::dnspack::CartesianCoordinate& rk::dnspack::CartesianCoordinate::scaling(const double num) noexcept
{
	x *= num;
	y *= num;
	z *= num;

	return *this;
}

inline double rk::dnspack::CartesianCoordinate::operator[](const int direction) const
{
	switch (direction)
	{
	case 0:
		return x;
		break;

	case 1:
		return y;
		break;

	case 2:
		return z;
		break;

	default:
		throw std::invalid_argument("CartesianCoordinate::operator[](const int direction) : direction is unsuitable.");
	}
}

inline double rk::dnspack::CartesianCoordinate::getNorm() const noexcept
{
	return sqrt(x * x + y * y + z * z);
}

inline double rk::dnspack::CartesianCoordinate::getNormSquare() const noexcept
{
	return (x * x + y * y + z * z);
}

namespace rk
{
	namespace dnspack
	{
		inline CartesianCoordinate operator+(CartesianCoordinate coordinate, const CartesianCoordinate& coordinate2)
		{
			coordinate.x += coordinate2.x;
			coordinate.y += coordinate2.y;
			coordinate.z += coordinate2.z;

			return coordinate;
		}

		inline CartesianCoordinate operator-(CartesianCoordinate coordinate, const CartesianCoordinate& coordinate2)
		{
			coordinate.x -= coordinate2.x;
			coordinate.y -= coordinate2.y;
			coordinate.z -= coordinate2.z;

			return coordinate;
		}

		inline CartesianCoordinate scaling(const double num, CartesianCoordinate coordinate)
		{
			coordinate.x *= num;
			coordinate.y *= num;
			coordinate.z *= num;

			return coordinate;
		}

		inline double innerProduct(const CartesianCoordinate& coordinate, const CartesianCoordinate& coordinate2)
		{
			return (coordinate.x * coordinate2.x + coordinate.y * coordinate2.y + coordinate.z * coordinate2.z);
		}

		inline CartesianCoordinate crossProduct(const CartesianCoordinate& coordinate, const CartesianCoordinate& coordinate2)
		{
			CartesianCoordinate crossProduct;

			crossProduct.x = coordinate.y * coordinate2.z - coordinate.z * coordinate2.y;
			crossProduct.y = coordinate.z * coordinate2.x - coordinate.x * coordinate2.z;
			crossProduct.z = coordinate.x * coordinate2.y - coordinate.y * coordinate2.x;

			return crossProduct;
		}
	}
}


#endif // !RK_DNSPACK_CARTESIANCOORDINATE_H
