#ifndef RK_DNSPACK_EXPLORESCHEDULE_H
#define RK_DNSPACK_EXPLORESCHEDULE_H

#include <mutex>
#include <string>
#include <map>

#include "PackingStructure.h"

namespace rk
{
	namespace dnspack
	{
		class ExploreSchedule
		{
		public:
			explicit ExploreSchedule(const std::string& resultDirectory);
			virtual ~ExploreSchedule();

		public:
			ExploreSchedule(const ExploreSchedule&) = default;
			ExploreSchedule(ExploreSchedule&&) noexcept = default;
			ExploreSchedule& operator=(const ExploreSchedule&) = default;
			ExploreSchedule& operator=(ExploreSchedule&&) noexcept = default;

		public:
			int getInitialSupplyingNumber() const noexcept;
			int getInitialDemandNumber() const noexcept;
			int getSupplyingNumber() const noexcept;  // The number of structures which have already been generated.
			int getDemandNumber() const noexcept;  // The number of structures planned currently to generate.

			double getMaxPackingFraction() const noexcept;
			const std::multimap<double, PackingStructure>& getDensePackingStructures() const noexcept;
			int getMaxPackingFractionStructureNumber() const noexcept;

			int getFullOptimizeStructureNumber() const noexcept;
			int getMaxTrialNumber() const noexcept; // The maximum number of structures to generate.
			int getThresholdTrialNumber() const noexcept;
			double getOutputMinimumPackingFraction() const noexcept;
			double getOutputMinimumZ() const noexcept;
			bool outputMdFile() const noexcept;
			int getDensePackingStructureNumber() const noexcept;
			int getThreadNumber() const noexcept;

			const std::string& getResultDirectory() const noexcept;
			const std::string& getResultDirectoryForDatFiles() const noexcept;
			const std::string& getResultDirectoryForCifFiles() const noexcept;

		public:
			void initialProduction() noexcept;
			void initialProduction(int& supplyingLabel) noexcept;
			void initialDemandReceive() noexcept;
			void production() noexcept;
			void production(int& supplyingLabel) noexcept;
			void demandReceive() noexcept;
			void productionStop() noexcept;

			void setDensePackingStructures(const PackingStructure& packingStructure);
			void appendMaxPackingFractionStructureNumber() noexcept;
			void renewMaxPackingFractionStructureNumber() noexcept;

		public:
			static void setMaxPackingFraction(const double maxPackingFraction);
			static void normalizeDensePackingStructures() noexcept;
			static void densePackingStructuresClear() noexcept;
			static void resetSchedule() noexcept;

		private:
			static int m_initialSupplyingNumber;
			static int m_initialDemandNumber;
			static int m_supplyingNumber;
			static int m_demandNumber;

			static double m_maxPackingFraction;
			static std::multimap<double, PackingStructure> m_densePackingStructures;
			static int m_maxPackingFractionStructureNumber;

			static std::mutex m_scheduleMutex;

		private:
			int m_fullOptimizeStructureNumber;
			int m_maxTrialNumber;
			int m_thresholdTrialNumber;
			
			double m_outputMinimumPackingFraction;
			double m_outputMinimumZ;
			bool m_outputMdFile;
			int m_densePackingStructureNumber;

			int m_threadNumber;

			std::string m_resultDirectory;
			std::string m_resultDirectoryForDatFiles;
			std::string m_resultDirectoryForCifFiles;
		};
	}
}


inline int rk::dnspack::ExploreSchedule::getInitialSupplyingNumber() const noexcept
{
	return m_initialSupplyingNumber;
}

inline int rk::dnspack::ExploreSchedule::getInitialDemandNumber() const noexcept
{
	return m_initialDemandNumber;
}

inline int rk::dnspack::ExploreSchedule::getSupplyingNumber() const noexcept
{
	return m_supplyingNumber;
}

inline int rk::dnspack::ExploreSchedule::getDemandNumber() const noexcept
{
	return m_demandNumber;
}

inline double rk::dnspack::ExploreSchedule::getMaxPackingFraction() const noexcept
{
	return m_maxPackingFraction;
}

inline const std::multimap<double, rk::dnspack::PackingStructure>& rk::dnspack::ExploreSchedule::getDensePackingStructures() const noexcept
{
	return m_densePackingStructures;
}

inline int rk::dnspack::ExploreSchedule::getMaxPackingFractionStructureNumber() const noexcept
{
	return m_maxPackingFractionStructureNumber;
}

inline int rk::dnspack::ExploreSchedule::getFullOptimizeStructureNumber() const noexcept
{
	return m_fullOptimizeStructureNumber;
}

inline int rk::dnspack::ExploreSchedule::getMaxTrialNumber() const noexcept
{
	return m_maxTrialNumber;
}

inline int rk::dnspack::ExploreSchedule::getThresholdTrialNumber() const noexcept
{
	return m_thresholdTrialNumber;
}

inline double rk::dnspack::ExploreSchedule::getOutputMinimumPackingFraction() const noexcept
{
	return m_outputMinimumPackingFraction;
}

inline double rk::dnspack::ExploreSchedule::getOutputMinimumZ() const noexcept
{
	return m_outputMinimumZ;
}

inline bool rk::dnspack::ExploreSchedule::outputMdFile() const noexcept
{
	return m_outputMdFile;
}

inline int rk::dnspack::ExploreSchedule::getDensePackingStructureNumber() const noexcept
{
	return m_densePackingStructureNumber;
}

inline int rk::dnspack::ExploreSchedule::getThreadNumber() const noexcept
{
	return m_threadNumber;
}

inline const std::string& rk::dnspack::ExploreSchedule::getResultDirectory() const noexcept
{
	return m_resultDirectory;
}

inline const std::string& rk::dnspack::ExploreSchedule::getResultDirectoryForDatFiles() const noexcept
{
	return m_resultDirectoryForDatFiles;
}

inline const std::string& rk::dnspack::ExploreSchedule::getResultDirectoryForCifFiles() const noexcept
{
	return m_resultDirectoryForCifFiles;
}

inline void rk::dnspack::ExploreSchedule::initialProduction() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_initialSupplyingNumber += 1;
}

inline void rk::dnspack::ExploreSchedule::initialProduction(int& initialSupplyingLabel) noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_initialSupplyingNumber += 1;
	initialSupplyingLabel = m_initialSupplyingNumber;
}

inline void rk::dnspack::ExploreSchedule::initialDemandReceive() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_initialDemandNumber = m_fullOptimizeStructureNumber;
}

inline void rk::dnspack::ExploreSchedule::production() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_supplyingNumber += 1;
}

inline void rk::dnspack::ExploreSchedule::production(int& supplyingLabel) noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_supplyingNumber += 1;
	supplyingLabel = m_supplyingNumber;
}

inline void rk::dnspack::ExploreSchedule::demandReceive() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	int number = m_supplyingNumber + m_thresholdTrialNumber;

	if (m_maxTrialNumber < number)
	{
		number = m_maxTrialNumber;
	}

	m_demandNumber = number;
}

inline void rk::dnspack::ExploreSchedule::productionStop() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_demandNumber = m_supplyingNumber;
}

inline void rk::dnspack::ExploreSchedule::setDensePackingStructures(const rk::dnspack::PackingStructure& packingStructure)
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	double packingFraction = packingStructure.getPackingFraction();

	if (m_densePackingStructures.size() < static_cast<std::size_t>(m_densePackingStructureNumber))
	{
		m_densePackingStructures.emplace(packingFraction, packingStructure);

		std::multimap<double, PackingStructure>::iterator iterPrevEnd = std::prev(m_densePackingStructures.end());
		m_maxPackingFraction = iterPrevEnd->first;
	}

	else if(m_densePackingStructures.size() == static_cast<std::size_t>(m_densePackingStructureNumber))
	{
		m_densePackingStructures.emplace(packingFraction, packingStructure);

		std::multimap<double, PackingStructure>::iterator iter = m_densePackingStructures.begin();
		m_densePackingStructures.erase(iter);

		std::multimap<double, PackingStructure>::iterator iterPrevEnd = std::prev(m_densePackingStructures.end());
		m_maxPackingFraction = iterPrevEnd->first;
	}

	else
	{
		throw std::logic_error("ExploreSchedule::m_densePackingStructures contains more elements than m_densePackingStructureNumber");
	}
}

inline void rk::dnspack::ExploreSchedule::appendMaxPackingFractionStructureNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_maxPackingFractionStructureNumber += 1;
}

inline void rk::dnspack::ExploreSchedule::renewMaxPackingFractionStructureNumber() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_maxPackingFractionStructureNumber = 1;
}

inline void rk::dnspack::ExploreSchedule::setMaxPackingFraction(const double maxPackingFraction)
{
	if (maxPackingFraction < 0.0)
	{
		throw std::logic_error("maxPackingFraction is less than one.");
	}

	else if (1.0 < maxPackingFraction)
	{
		throw std::logic_error("maxPackingFraction is more than one.");
	}

	else
	{
		m_maxPackingFraction = maxPackingFraction;
	}
}

inline void rk::dnspack::ExploreSchedule::normalizeDensePackingStructures() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	double thresholdPackingFraction = 0.0;
	{
		if (!m_densePackingStructures.empty())
		{
			std::multimap<double, PackingStructure>::iterator iterPrevEnd = std::prev(m_densePackingStructures.end());
			thresholdPackingFraction = iterPrevEnd->first - 0.001;
		}
	}

	for (std::multimap<double, PackingStructure>::iterator iter = m_densePackingStructures.begin(); iter != m_densePackingStructures.end();)
	{
		if (iter->first < thresholdPackingFraction)
		{
			iter = m_densePackingStructures.erase(iter);
		}

		else
		{
			++iter;
		}
	}
}

inline void rk::dnspack::ExploreSchedule::densePackingStructuresClear() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_densePackingStructures.clear();
}

inline void rk::dnspack::ExploreSchedule::resetSchedule() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	m_initialSupplyingNumber = 0;
	m_initialDemandNumber = 0;
	m_supplyingNumber = 0;
	m_demandNumber = 0;

	m_maxPackingFraction = 0.0;
	m_densePackingStructures.clear();
	m_maxPackingFractionStructureNumber = 0;
}

#endif // !RK_DNSPACK_EXPLORESCHEDULE_H
