#ifndef RK_DNSPACK_ITERATIVEBALANCEOPTIMIZE_H
#define RK_DNSPACK_ITERATIVEBALANCEOPTIMIZE_H

#include <valarray>

#include "SphereLabel.h"
#include "PackingStructure.h"
#include "MakeMdFile.h"


namespace rk
{
	namespace dnspack
	{
		class IterativeBalanceOptimize
		{
		public:
			IterativeBalanceOptimize(PackingStructure& packingStructure, const int maxStepNumber, const int neighborUpdateInterval, const double decreaseFactor, const double maxDisplacement);
			virtual ~IterativeBalanceOptimize();

		public:
			bool operator()();
			bool operator()(MakeMdFile& makeMdFile);

		private:
			bool setNeighborSpheres();
			void calculateGradient();
			void optimize();

			double potential(const double z) const noexcept;
			double force(const double z) const noexcept;

		private:
			PackingStructure& m_packingStructure;

			const int m_spheresNumber;
			const int m_maxStepNumber;
			const int m_neighborUpdateInterval;

			std::vector<std::pair<int, SphereLabel>> m_neighborSpheres{}; // baseIndex <= sphereIndex̏ꍇ̂
			std::valarray<double> m_gradientVector{};

			double m_maxDisplacement;
			double m_decreaseFactor;
			const double m_pressure{ 0.1 };

		private:
			IterativeBalanceOptimize(const IterativeBalanceOptimize&) = delete;
			IterativeBalanceOptimize(IterativeBalanceOptimize&&) noexcept = delete;
			IterativeBalanceOptimize& operator=(const IterativeBalanceOptimize&) = delete;
			IterativeBalanceOptimize& operator=(IterativeBalanceOptimize&&) noexcept = delete;
		};
	}
}


inline bool rk::dnspack::IterativeBalanceOptimize::operator()()
{
	int repNumber = m_maxStepNumber / m_neighborUpdateInterval;
	int repSurplus = m_maxStepNumber % m_neighborUpdateInterval;

	for (int rep = 0; rep < repNumber; ++rep)
	{
		if (setNeighborSpheres())
		{
			for (int step = 0; step < m_neighborUpdateInterval; ++step)
			{
				calculateGradient();
				optimize();

				m_maxDisplacement *= m_decreaseFactor;
			}

			if (m_packingStructure.getBravaisLattice().getVolume() < 0)
			{
				return false;
			}
		}

		else
		{
			return false;
		}
	}

	{
		if (setNeighborSpheres())
		{
			for (int step = 0; step < repSurplus; ++step)
			{
				calculateGradient();
				optimize();

				m_maxDisplacement *= m_decreaseFactor;
			}

			if (m_packingStructure.getBravaisLattice().getVolume() < 0)
			{
				return false;
			}
		}

		else
		{
			return false;
		}
	}

	return true;
}

inline bool rk::dnspack::IterativeBalanceOptimize::operator()(rk::dnspack::MakeMdFile& makeMdFile)
{
	makeMdFile.takePicture("IterativeBalanceOptimize");

	int repNumber = m_maxStepNumber / m_neighborUpdateInterval;
	int repSurplus = m_maxStepNumber % m_neighborUpdateInterval;

	for (int rep = 0; rep < repNumber; ++rep)
	{
		if (setNeighborSpheres())
		{
			for (int step = 0; step < m_neighborUpdateInterval; ++step)
			{
				calculateGradient();
				optimize();

				m_maxDisplacement *= m_decreaseFactor;
			}

			if (m_packingStructure.getBravaisLattice().getVolume() < 0)
			{
				return false;
			}

			makeMdFile.takePicture();
		}

		else
		{
			return false;
		}
	}

	{
		if (setNeighborSpheres())
		{
			for (int step = 0; step < repSurplus; ++step)
			{
				calculateGradient();
				optimize();

				m_maxDisplacement *= m_decreaseFactor;
			}

			if (m_packingStructure.getBravaisLattice().getVolume() < 0)
			{
				return false;
			}
		}

		else
		{
			return false;
		}
	}

	makeMdFile.takePicture();
	return true;
}

inline double rk::dnspack::IterativeBalanceOptimize::potential(const double z) const noexcept
{
	if (z < 0)
	{
		return (-1.0 * z);
	}

	else
	{
		return (0);
	}
}

inline double rk::dnspack::IterativeBalanceOptimize::force(const double z) const noexcept
{
	if (z < 0)
	{
		return (-1.0);
	}

	else
	{
		return (0);
	}
}



#endif // !RK_DNSPACK_ITERATIVEBALANCEOPTIMIZE_H
