#ifndef RK_DNSPACK_REOPTIMIZESCHEDULE_H
#define RK_DNSPACK_REOPTIMIZESCHEDULE_H

#include <mutex>
#include <string>

#include "PackingStructure.h"

namespace rk
{
	namespace dnspack
	{
		class ReOptimizeSchedule
		{
		public:
			ReOptimizeSchedule(const PackingStructure& seedStructures, const std::string& densestPackingsDirectory);
			virtual ~ReOptimizeSchedule();

		public:
			ReOptimizeSchedule(const ReOptimizeSchedule&) = default;
			ReOptimizeSchedule(ReOptimizeSchedule&&) noexcept = default;
			ReOptimizeSchedule& operator=(const ReOptimizeSchedule&) = default;
			ReOptimizeSchedule& operator=(ReOptimizeSchedule&&) noexcept = default;

		public:
			int getThreadNumber() const noexcept;

			const PackingStructure& getSeedStructure() const noexcept;
			int getReOptimizeRepetitionNumber() const noexcept; // The maximum number of structures to generate.
			double getOutputMinimumZ() const noexcept;

			int getSupplyingNumber() const noexcept;  // The number of structures which have already been generated.
			int getDemandNumber() const noexcept; // The number of structures planned currently to generate.
			double getMaxPackingFraction() const noexcept;

			const std::string& getDensestPackingsDirectory() const noexcept;
			const std::string& getDensestPackingsDirectoryForDatFiles() const noexcept;
			const std::string& getDensestPackingsDirectoryForCifFiles() const noexcept;

		public:
			void reOptimization() noexcept;
			void reOptimization(int& supplyingLabel) noexcept;
			void demandReceive() noexcept;
			void setMaxPackingFraction(const double& maxPackingFraction);

			static void resetSchedule() noexcept;

		private:
			static int m_supplyingNumber;
			static int m_demandNumber;
			static double m_maxPackingFraction;

			static std::mutex m_scheduleMutex;

		private:
			int m_threadNumber;

			PackingStructure m_seedStructure;
			int m_reOptimizeRepetitionNumber;
			double m_outputMinimumZ;

			std::string m_densestPackingsDirectory;
			std::string m_densestPackingsDirectoryForDatFiles;
			std::string m_densestPackingsDirectoryForCifFiles;
		};
	}
}


inline const rk::dnspack::PackingStructure& rk::dnspack::ReOptimizeSchedule::getSeedStructure() const noexcept
{
	return m_seedStructure;
}

inline int rk::dnspack::ReOptimizeSchedule::getReOptimizeRepetitionNumber() const noexcept
{
	return m_reOptimizeRepetitionNumber;
}

inline int rk::dnspack::ReOptimizeSchedule::getThreadNumber() const noexcept
{
	return m_threadNumber;
}

inline double rk::dnspack::ReOptimizeSchedule::getOutputMinimumZ() const noexcept
{
	return m_outputMinimumZ;
}

inline int rk::dnspack::ReOptimizeSchedule::getSupplyingNumber() const noexcept
{
	return m_supplyingNumber;
}

inline int rk::dnspack::ReOptimizeSchedule::getDemandNumber() const noexcept
{
	return m_demandNumber;
}

inline double rk::dnspack::ReOptimizeSchedule::getMaxPackingFraction() const noexcept
{
	return m_maxPackingFraction;
}

inline const std::string& rk::dnspack::ReOptimizeSchedule::getDensestPackingsDirectory() const noexcept
{
	return m_densestPackingsDirectory;
}

inline const std::string& rk::dnspack::ReOptimizeSchedule::getDensestPackingsDirectoryForDatFiles() const noexcept
{
	return m_densestPackingsDirectoryForDatFiles;
}

inline const std::string& rk::dnspack::ReOptimizeSchedule::getDensestPackingsDirectoryForCifFiles() const noexcept
{
	return m_densestPackingsDirectoryForCifFiles;
}

inline void rk::dnspack::ReOptimizeSchedule::reOptimization() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_supplyingNumber += 1;
}

inline void rk::dnspack::ReOptimizeSchedule::reOptimization(int& supplyingLabel) noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };
	m_supplyingNumber += 1;
	supplyingLabel = m_supplyingNumber;
}

inline void rk::dnspack::ReOptimizeSchedule::demandReceive() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	m_demandNumber = m_reOptimizeRepetitionNumber;
}

inline void rk::dnspack::ReOptimizeSchedule::setMaxPackingFraction(const double& maxPackingFraction)
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	if (m_maxPackingFraction < maxPackingFraction)
	{
		m_maxPackingFraction = maxPackingFraction;
	}
}

inline void rk::dnspack::ReOptimizeSchedule::resetSchedule() noexcept
{
	std::lock_guard<std::mutex> lock{ m_scheduleMutex };

	m_supplyingNumber = 0;
	m_demandNumber = 0;
	m_maxPackingFraction = 0.0;
}


#endif // !RK_DNSPACK_REOPTIMIZESCHEDULE_H
