#define _USE_MATH_DEFINES

#include "MakeCifFile.h"

#include <cmath>
#include <map>
#include <sstream>
#include <fstream>
#include <stdexcept>

#include "Clock.h"

using namespace rk::dnspack;


std::mutex MakeCifFile::m_cifMutex;


MakeCifFile::MakeCifFile()
{
}

MakeCifFile::~MakeCifFile()
{
}

void MakeCifFile::operator()(const std::string& fileName, const PackingStructure& packingStructure)
{
   BasisComposition basisComposition(packingStructure.getBasisComposition());

// ********************************************************************************************************************************************************************
    // initialize sphere names

    std::size_t sphereKindNumber = packingStructure.getSphereKindNumber();

    std::vector<std::string> sphereNames(6);
    {
        sphereNames[0] = "H";
        sphereNames[1] = "Na";
        sphereNames[2] = "Sc";
        sphereNames[3] = "Y";
        sphereNames[4] = "La";
        sphereNames[5] = "Hf";
    }

    std::vector<double> radii;
    {
        for (const auto& radius : basisComposition.getRadii())
        {
            radii.push_back(radius);
        }
    }

    if (6 < radii.size())
    {
        throw std::runtime_error("The kind of sphere is larger than 6.");
    }

    std::map<double, std::string> radiiAndName;
    {
        std::vector<std::string>::const_iterator nameIter = sphereNames.begin();

        for (const auto& radius : radii)
        {
            radiiAndName.emplace(radius, *nameIter);
            ++nameIter;
        }
    }

    // initialize sphere names
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
    // making structural information for a CIF file

    double cellVolume = packingStructure.getBravaisLattice().getVolume();
    const BravaisLattice& bravaisLattice = packingStructure.getBravaisLattice();

    double latticeLength[3]{ 0.0,0.0,0.0 };
    {
        latticeLength[0] = bravaisLattice[0].getNorm();
        latticeLength[1] = bravaisLattice[1].getNorm();
        latticeLength[2] = bravaisLattice[2].getNorm();
    }

    double alpha = 0;
    double beta = 0;
    double gamma = 0;

    {
        double dotProduct = innerProduct(bravaisLattice[1], bravaisLattice[2]);

        if (fabs(dotProduct) < 1.0e-14)
        {
            alpha = 90.0;
        }
        else
        {
            alpha = acos(dotProduct / (latticeLength[1] * latticeLength[2])) / M_PI * 180.0;
        }
    }

    {
        double dotProduct = innerProduct(bravaisLattice[2], bravaisLattice[0]);

        if (fabs(dotProduct) < 1.0e-14)
        {
            beta = 90.0;
        }
        else
        {
            beta = acos(dotProduct / (latticeLength[2] * latticeLength[0])) / M_PI * 180.0;
        }
    }

    {
        double dotProduct = innerProduct(bravaisLattice[0], bravaisLattice[1]);

        if (fabs(dotProduct) < 1.0e-14)
        {
            gamma = 90.0;
        }
        else
        {
            gamma = acos(dotProduct / (latticeLength[0] * latticeLength[1])) / M_PI * 180.0;
        }
    }

    std::lock_guard<std::mutex> lock{ m_cifMutex };

    std::stringstream textStream;
    textStream.precision(12);

    textStream << "data_" << fileName << "\n";
    textStream << "_audit_creation_date              " << Clock::getNowTime() << "\n";
    textStream << "_audit_creation_method            'SAMLAI'\n";
    textStream << "_symmetry_space_group_name_H-M    'P1'\n";
    textStream << "_symmetry_Int_Tables_number       1\n";
    textStream << "_symmetry_cell_setting            triclinic\n";
    textStream << "loop_\n";
    textStream << "_symmetry_equiv_pos_as_xyz\n";
    textStream << "  x,y,z\n";

    textStream << "_cell_length_a               " << latticeLength[0] << "\n";
    textStream << "_cell_length_b               " << latticeLength[1] << "\n";
    textStream << "_cell_length_c               " << latticeLength[2] << "\n";
    textStream << "_cell_angle_alpha            " << alpha << "\n";
    textStream << "_cell_angle_beta             " << beta << "\n";
    textStream << "_cell_angle_gamma            " << gamma << "\n";

    textStream << "loop_\n";
    
    textStream << "_atom_site_label\n";
    textStream << "_atom_site_type_symbol\n";
    textStream << "_atom_site_fract_x\n";
    textStream << "_atom_site_fract_y\n";
    textStream << "_atom_site_fract_z\n";
    textStream << "_atom_site_Uiso_or_equiv\n";
    textStream << "_atom_site_adp_type\n";
    textStream << "_atom_site_occupancy\n";

    {
        int num = 1;
        double zero = 0.0;
        double one = 1.0;

        for (auto sphere : packingStructure.getBasis())
        {
            sphere.refineFractionalCoordinate();

            textStream << "A" << num << "   " << radiiAndName.at(sphere.getRadius()) << "  ";
            textStream << sphere.getFractionalCoordinate().a << "  " << sphere.getFractionalCoordinate().b << "  " << sphere.getFractionalCoordinate().c << "   ";
            textStream << zero << "   Uiso   " << one << "\n";

            ++num;
        }
    }

    // making structural information for a CIF file
// *******************************************************************************************************************************************************************

// *******************************************************************************************************************************************************************
    // making of a CIF file

    std::ofstream outputFileStream;
    outputFileStream.open(fileName, std::ios_base::out);

    if (!outputFileStream)
    {
        throw std::invalid_argument("couldn't open dat file.");
    }

    outputFileStream << textStream.str();
    outputFileStream.close();

    // making of a CIF file
// *******************************************************************************************************************************************************************

}
