#include "MakeDatFile.h"

#include <map>
#include <sstream>
#include <stdexcept>

using namespace rk::dnspack;


std::mutex MakeDatFile::m_datMutex;


MakeDatFile::MakeDatFile()
	: m_outputString{}
{
}

MakeDatFile::~MakeDatFile()
{
}

void MakeDatFile::writePackingInfo(const std::string& fileName, const PackingStructure& packingStructure)
{
	BasisComposition basisComposition(packingStructure.getBasisComposition());

// ********************************************************************************************************************************************************************
	// initialize sphere names

	std::size_t sphereKindNumber = packingStructure.getSphereKindNumber();

	std::vector<std::string> sphereNames(6);
	{
		sphereNames[0] = "H";
		sphereNames[1] = "Na";
		sphereNames[2] = "Sc";
		sphereNames[3] = "Y";
		sphereNames[4] = "La";
		sphereNames[5] = "Hf";
	}

	std::vector<double> radii;
	{
		for (const auto& radius : basisComposition.getRadii())
		{
			radii.push_back(radius);
		}
	}

	if (6 < radii.size())
	{
		m_outputString.clear();
		throw std::runtime_error("The kind of sphere is larger than 6.");
	}

	std::map<double, std::string> radiiAndName;
	{
		std::vector<std::string>::const_iterator nameIter = sphereNames.begin();

		for (const auto& radius : radii)
		{
			radiiAndName.emplace(radius, *nameIter);
			++nameIter;
		}
	}

	// initialize sphere names
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// making structural information for a DAT file

	std::stringstream textStream;
	textStream.precision(12);

	textStream << "#\n" << "# Spheres Infomation : The name and the radius of each sphere\n" << "#\n\n\n";
	textStream << "<Definition.of.Sphere.Species\n";

	for (std::size_t label = 0; label < sphereKindNumber; ++label)
	{
		textStream << " " << sphereNames[label] << "    " << radii[label] << "\n";
	}
	textStream << "Definition.of.Sphere.Species>\n\n";

	textStream << "Atoms.Number   " << packingStructure.getSpheresNumber() << "\n";
	textStream << "<Number.Of.Spheres.Per.UnitCell\n";
	{
		std::size_t label = 0;

		for (const auto& number : basisComposition.getComposition())
		{
			textStream << " " << sphereNames[label] << "    " << number.second << "\n";
			++label;
		}
	}
	textStream << "Number.Of.Spheres.Per.UnitCell>\n\n\n";


	textStream << "#\n" << "# Structural Infomation\n" << "#\n\n\n";

	textStream << "Atoms.SpeciesAndCoordinates.Unit  FRAC\n<Atoms.SpeciesAndCoordinates" << "\n";
	{
		int num = 1;

		for (const auto& sphere : packingStructure.getBasis())
		{
			textStream << "  " << num << "  " << radiiAndName.at(sphere.getRadius()) << "  ";
			textStream << sphere.getFractionalCoordinate().a << "  " << sphere.getFractionalCoordinate().b << "  " << sphere.getFractionalCoordinate().c << "\n";

			++num;
		}
	}
	textStream << "Atoms.SpeciesAndCoordinates>" << "\n\n\n";

	const BravaisLattice& bravaisLattice = packingStructure.getBravaisLattice();

	textStream << "Atoms.UnitVectors.Unit   Ang\n<Atoms.UnitVectors\n";
	textStream << "  " << bravaisLattice[0].x << "  " << bravaisLattice[0].y << "  " << bravaisLattice[0].z << "\n";
	textStream << "  " << bravaisLattice[1].x << "  " << bravaisLattice[1].y << "  " << bravaisLattice[1].z << "\n";
	textStream << "  " << bravaisLattice[2].x << "  " << bravaisLattice[2].y << "  " << bravaisLattice[2].z << "\n";
	textStream << "Atoms.UnitVectors>" << "\n\n\n";

	double overlapVolume = 0.0;
	double minimumZ = 0.0;
	{
		if (packingStructure.overlapExistence(overlapVolume, minimumZ))
		{
			textStream << "<PackingInformation\n";
			textStream << "packing fraction  :  " << packingStructure.getPackingFraction() << "\n";
			textStream << "overlap volume    :  " << " Exist \n";
			textStream << "PackingInformation>\n\n\n";
		}

		else
		{
			textStream << "<PackingInformation\n";
			textStream << "packing fraction  :   " << packingStructure.getPackingFraction() << "\n";
			textStream << "overlap volume    :   " << overlapVolume << "\n";
			textStream << "minimumZ          :   " << minimumZ << "\n";
			textStream << "PackingInformation>\n\n\n";
		}
	}

	// making structural information for a DAT file
// *******************************************************************************************************************************************************************

	m_outputString += textStream.str();
}
