#include "MakeMdFile.h"

#include <map>
#include <sstream>
#include <stdexcept>

#include "Clock.h"

using namespace rk::dnspack;

std::mutex MakeMdFile::m_mdMutex;


MakeMdFile::MakeMdFile(PackingStructure& packingStructure)
	: m_packingStructure(packingStructure)
	, m_spheresNumber{ static_cast<int>(packingStructure.getSpheresNumber()) }
	, m_moveRecord{}
	, m_stepNumber{ 0 }
	, m_radiiAndName()
{
	BasisComposition basisComposition = m_packingStructure.getBasisComposition();

	if (6 < basisComposition.getSphereKindNumber())
	{
		throw std::runtime_error("The sphere kind number is more than 6.");
	}
	
	else
	{
		std::vector<std::string> sphereNames(6);
		{
			sphereNames[0] = "H";
			sphereNames[1] = "Na";
			sphereNames[2] = "Sc";
			sphereNames[3] = "Y";
			sphereNames[4] = "La";
			sphereNames[5] = "Hf";
		}

		std::vector<std::string>::const_iterator nameIter = sphereNames.begin();


		for (const auto& radiiAndNum : basisComposition.getComposition())
		{
			m_radiiAndName.emplace(radiiAndNum.first, *nameIter);
			++nameIter;
		}
	}
}

MakeMdFile::~MakeMdFile()
{
}

void MakeMdFile::takePicture()
{
	m_stepNumber += 1;

	double packingFraction = m_packingStructure.getPackingFraction();

	m_moveRecord += std::to_string(m_spheresNumber);
	m_moveRecord += "\nstep= ";
	m_moveRecord += std::to_string(m_stepNumber);
	m_moveRecord += "   PackingFraction= ";
	m_moveRecord += std::to_string(packingFraction);
	m_moveRecord += "   Cell_Vectors=  ";
	
	const BravaisLattice& bravaisLattice = m_packingStructure.getBravaisLattice();
	m_moveRecord += std::to_string(bravaisLattice[0].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[0].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[0].z);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].z);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].z);
	m_moveRecord += "\n";
		
	const std::vector<Sphere>& basis = m_packingStructure.getBasis();

	std::string zeroColumns = " ";
	{
		std::string zero = "  0.00000";

		for (int label = 0; label < 10; ++label)
		{
			zeroColumns += zero;
		}
	}

	for (int label = 0; label < m_spheresNumber; ++label)
	{
		Sphere sphere = basis.at(static_cast<std::size_t>(label));

		m_moveRecord += " ";
		m_moveRecord += m_radiiAndName.at(sphere.getRadius());
		m_moveRecord += "    ";

		CartesianCoordinate cartesianCoordinate;

		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().a, bravaisLattice[0]);
		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().b, bravaisLattice[1]);
		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().c, bravaisLattice[2]);

		m_moveRecord += std::to_string(cartesianCoordinate.x);
		m_moveRecord += "  ";
		m_moveRecord += std::to_string(cartesianCoordinate.y);
		m_moveRecord += "  ";
		m_moveRecord += std::to_string(cartesianCoordinate.z);
		m_moveRecord += zeroColumns;
		m_moveRecord += "\n";
	}
}

void MakeMdFile::takePicture(const std::string& message)
{
	m_stepNumber += 1;

	double packingFraction = m_packingStructure.getPackingFraction();

	m_moveRecord += std::to_string(m_spheresNumber);
	m_moveRecord += "\nstep= ";
	m_moveRecord += std::to_string(m_stepNumber);
	m_moveRecord += "   PackingFraction= ";
	m_moveRecord += std::to_string(packingFraction);
	m_moveRecord += "   ";
	m_moveRecord += message;
	m_moveRecord += "   Cell_Vectors=  ";

	const BravaisLattice& bravaisLattice = m_packingStructure.getBravaisLattice();
	m_moveRecord += std::to_string(bravaisLattice[0].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[0].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[0].z);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[1].z);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].x);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].y);
	m_moveRecord += "  ";
	m_moveRecord += std::to_string(bravaisLattice[2].z);
	m_moveRecord += "\n";

	const std::vector<Sphere>& basis = m_packingStructure.getBasis();

	std::string zeroColumns = " ";
	{
		std::string zero = "  0.00000";

		for (int label = 0; label < 10; ++label)
		{
			zeroColumns += zero;
		}
	}

	for (int label = 0; label < m_spheresNumber; ++label)
	{
		Sphere sphere = basis.at(static_cast<std::size_t>(label));

		m_moveRecord += " ";
		m_moveRecord += m_radiiAndName.at(sphere.getRadius());
		m_moveRecord += "    ";

		CartesianCoordinate cartesianCoordinate;

		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().a, bravaisLattice[0]);
		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().b, bravaisLattice[1]);
		cartesianCoordinate += scaling(sphere.getFractionalCoordinate().c, bravaisLattice[2]);

		m_moveRecord += std::to_string(cartesianCoordinate.x);
		m_moveRecord += "  ";
		m_moveRecord += std::to_string(cartesianCoordinate.y);
		m_moveRecord += "  ";
		m_moveRecord += std::to_string(cartesianCoordinate.z);
		m_moveRecord += zeroColumns;
		m_moveRecord += "\n";
	}
}
