#include "PackingFactory.h"

#include <sstream>
#include <random>
#include <thread>
#include <mutex>

#include "PilingUpGenerate.h"
#include "Annealing.h"
#include "IterativeBalanceOptimize.h"

#include "MakeDatFile.h"
#include "MakeCifFile.h"
#include "MakeMdFile.h"
#include "MakeDirectory.h"

using namespace rk::dnspack;


PackingFactory::PackingFactory(const BasisComposition& basisComposition, const ExploreSchedule& exploreSchedule)
	: m_basisComposition(basisComposition)
	, m_exploreSchedule(exploreSchedule)
	, m_calculationAnalysis()
	, m_reportCurrentSituation(exploreSchedule.getResultDirectory() + "currentSituation.txt")
{
	EnvironmentVariables environmentVariables;
	m_exploreVariables = environmentVariables.getExploreVariables();
}

PackingFactory::~PackingFactory()
{
}

void PackingFactory::operator()()
{
	const std::thread::id myThreadID{ std::this_thread::get_id() };

	{
		std::stringstream reportMessage;
		reportMessage << "Thread ID  =  " << myThreadID << "  :  Begin to produce packing structures. \n";
		m_reportCurrentSituation.report(reportMessage.str());
	}

	if (m_exploreSchedule.outputMdFile())
	{
		executeWithMd();
	}

	else
	{
		execute();
	}
}

void PackingFactory::execute()
{
// *****************************************************************************************************************************************************************************************
	// full optimization in beginning

	while (m_exploreSchedule.getInitialSupplyingNumber() < m_exploreSchedule.getInitialDemandNumber())
	{
		int supplyingLabel = 0;
		m_exploreSchedule.initialProduction(supplyingLabel);

		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> annealingDist(m_exploreVariables.getAnnealingMaxDisplacementLower(), m_exploreVariables.getAnnealingMaxDisplacementUpper());
		std::uniform_real_distribution<double> packingOptimizeDist(m_exploreVariables.getPackingOptimizeMaxDisplacementLower(), m_exploreVariables.getPackingOptimizeMaxDisplacementUpper());


		PilingUpGenerate pilingUpGenerate(m_basisComposition);
		PackingStructure packingStructure = pilingUpGenerate();

		double annealingMaxDisplacement = annealingDist(engine);
		Annealing annealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), annealingMaxDisplacement);

		if (!annealing())
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double secondAnnealingMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize secondAnnealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), 0.9999, secondAnnealingMaxDisplacement);

		if (!secondAnnealing())
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double initialPackingOptimizeMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize initialPackingOptimize(packingStructure, m_exploreVariables.getInitialPackingOptimizeIterationNumber(), m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getInitialPackingOptimizeDecreaseFactor(), initialPackingOptimizeMaxDisplacement);

		if (!initialPackingOptimize())
		{
			m_calculationAnalysis.appendInitialPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		const double smallAnnealingPackingFraction = packingStructure.getPackingFraction();
		m_calculationAnalysis.appendInitialPackingOptimizePackingFractionsDistribution(smallAnnealingPackingFraction);

		double packingOptimizeMaxDisplacement = packingOptimizeDist(engine);
		IterativeBalanceOptimize spherePackingOptimize(packingStructure, m_exploreVariables.getPackingOptimizeIterationNumber(), m_exploreVariables.getPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getPackingOptimizeDecreaseFactor(), packingOptimizeMaxDisplacement);

		if (!spherePackingOptimize())
		{
			m_calculationAnalysis.appendPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
				continue;
			}

			else
			{
				if (minimumZ < m_exploreSchedule.getOutputMinimumZ())
				{
					m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
					continue;
				}

				else
				{
					const double packingFraction = packingStructure.getPackingFraction();

					m_calculationAnalysis.appendPackingFractionsError(smallAnnealingPackingFraction, packingFraction);
					m_calculationAnalysis.appendPackingOptimizePackingFractionsDistribution(packingFraction);

					if ((packingFraction - m_exploreSchedule.getMaxPackingFraction()) < -1.0e-5)
					{
						continue;
					}

					else if (-1.0e-5 < (packingFraction - m_exploreSchedule.getMaxPackingFraction()) && (packingFraction - m_exploreSchedule.getMaxPackingFraction()) < 1.0e-5)
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);

						std::stringstream message;
						message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find the putative densest packing fraction again of " << packingFraction << "\n";
						m_reportCurrentSituation.report(message.str());

						continue;
					}

					else
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.normalizeDensePackingStructures();

						if (packingFraction < m_exploreSchedule.getOutputMinimumPackingFraction())
						{
							std::stringstream message;
							message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());

							continue;
						}

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << supplyingLabel;

						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
						optimizeMessage << "Annealing.Maximum.Displacement                                     " << annealingMaxDisplacement << "\n";
						optimizeMessage << "Annealing.Iteration.Number                                         " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "Annealing.Interval.To.Update.The.Neighboringlist                   " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "SecondAnnealing.Maximum.Displacement                               " << secondAnnealingMaxDisplacement << "\n";
						optimizeMessage << "SecondAnnealing.Iteration.Number                                   " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "SecondAnnealing.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "InitialPackingOptimize.Maximum.Displacement                        " << initialPackingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "InitialPackingOptimize.Decrease.Factor                             " << m_exploreVariables.getInitialPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "InitialPackingOptimize.Iteration.Number                            " << m_exploreVariables.getInitialPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "InitialPackingOptimize.Interval.To.Update.The.Neighboringlist      " << m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval() << "\n";

						optimizeMessage << "PackingOptimize.Maximum.Displacement                               " << packingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "PackingOptimize.Decrease.Factor                                    " << m_exploreVariables.getPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "PackingOptimize.Iteration.Number                                   " << m_exploreVariables.getPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "PackingOptimize.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getPackingOptimizeNeighborUpdateInterval() << "\n";


						std::string datFileName = m_exploreSchedule.getResultDirectoryForDatFiles() + std::to_string(packingFraction) + ".dat";
						std::string cifFileName = m_exploreSchedule.getResultDirectoryForCifFiles() + std::to_string(packingFraction) + ".cif";

						MakeDatFile makeDatFile;
						makeDatFile(datFileName, packingStructure, optimizeMessage.str());

						MakeCifFile makeCifFile;
						makeCifFile(cifFileName, packingStructure);

						{
							std::stringstream message;
							message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());
						}
					}
				}
			}
		}
	}

	// full optimization in beginning
// ***************************************************************************************************************************************************************************************

// *****************************************************************************************************************************************************************************************
	// main search in the densest packings

	m_calculationAnalysis.setThresholdPackingFraction();

	while (m_exploreSchedule.getSupplyingNumber() < m_exploreSchedule.getDemandNumber())
	{
		int supplyingLabel = 0;
		m_exploreSchedule.production(supplyingLabel);

		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> annealingDist(m_exploreVariables.getAnnealingMaxDisplacementLower(), m_exploreVariables.getAnnealingMaxDisplacementUpper());
		std::uniform_real_distribution<double> packingOptimizeDist(m_exploreVariables.getPackingOptimizeMaxDisplacementLower(), m_exploreVariables.getPackingOptimizeMaxDisplacementUpper());


		PilingUpGenerate pilingUpGenerate(m_basisComposition);
		PackingStructure packingStructure = pilingUpGenerate();

		double annealingMaxDisplacement = annealingDist(engine);
		Annealing annealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), annealingMaxDisplacement);

		if (!annealing())
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double secondAnnealingMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize secondAnnealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), 0.9999, secondAnnealingMaxDisplacement);

		if (!secondAnnealing())
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double initialPackingOptimizeMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize initialPackingOptimize(packingStructure, m_exploreVariables.getInitialPackingOptimizeIterationNumber(), m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getInitialPackingOptimizeDecreaseFactor(), initialPackingOptimizeMaxDisplacement);

		if (!initialPackingOptimize())
		{
			m_calculationAnalysis.appendInitialPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		const double smallAnnealingPackingFraction = packingStructure.getPackingFraction();
		m_calculationAnalysis.appendInitialPackingOptimizePackingFractionsDistribution(smallAnnealingPackingFraction);

		if (smallAnnealingPackingFraction < m_calculationAnalysis.getThresholdPackingFraction())
		{
			m_calculationAnalysis.appendLowPackingFractionRejectNumber();
			continue;
		}

		double packingOptimizeMaxDisplacement = packingOptimizeDist(engine);
		IterativeBalanceOptimize spherePackingOptimize(packingStructure, m_exploreVariables.getPackingOptimizeIterationNumber(), m_exploreVariables.getPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getPackingOptimizeDecreaseFactor(), packingOptimizeMaxDisplacement);

		if (!spherePackingOptimize())
		{
			m_calculationAnalysis.appendPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
				continue;
			}

			else
			{
				if (minimumZ < m_exploreSchedule.getOutputMinimumZ())
				{
					m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
					continue;
				}

				else
				{
					const double packingFraction = packingStructure.getPackingFraction();

					m_calculationAnalysis.appendPackingFractionsError(smallAnnealingPackingFraction, packingFraction);
					m_calculationAnalysis.appendPackingOptimizePackingFractionsDistribution(packingFraction);
					m_calculationAnalysis.setThresholdPackingFraction();

					if ((packingFraction - m_exploreSchedule.getMaxPackingFraction()) < -1.0e-5)
					{
						continue;
					}

					else if (-1.0e-5 < (packingFraction - m_exploreSchedule.getMaxPackingFraction()) && (packingFraction - m_exploreSchedule.getMaxPackingFraction()) < 1.0e-5)
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.appendMaxPackingFractionStructureNumber();

						std::stringstream message;
						message << "Trial Label  =  " << supplyingLabel << "  :  Find the putative densest packing fraction again of " << packingFraction << "\n";
						m_reportCurrentSituation.report(message.str());

						continue;
					}

					else
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.normalizeDensePackingStructures();
						m_exploreSchedule.demandReceive();

						m_exploreSchedule.renewMaxPackingFractionStructureNumber();

						if (packingFraction < m_exploreSchedule.getOutputMinimumPackingFraction())
						{
							std::stringstream message;
							message << "Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());

							continue;
						}

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << supplyingLabel;

						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
						optimizeMessage << "Annealing.Maximum.Displacement                                     " << annealingMaxDisplacement << "\n";
						optimizeMessage << "Annealing.Iteration.Number                                         " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "Annealing.Interval.To.Update.The.Neighboringlist                   " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "SecondAnnealing.Maximum.Displacement                               " << secondAnnealingMaxDisplacement << "\n";
						optimizeMessage << "SecondAnnealing.Iteration.Number                                   " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "SecondAnnealing.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "InitialPackingOptimize.Maximum.Displacement                        " << initialPackingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "InitialPackingOptimize.Decrease.Factor                             " << m_exploreVariables.getInitialPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "InitialPackingOptimize.Iteration.Number                            " << m_exploreVariables.getInitialPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "InitialPackingOptimize.Interval.To.Update.The.Neighboringlist      " << m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval() << "\n";

						optimizeMessage << "PackingOptimize.Maximum.Displacement                               " << packingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "PackingOptimize.Decrease.Factor                                    " << m_exploreVariables.getPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "PackingOptimize.Iteration.Number                                   " << m_exploreVariables.getPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "PackingOptimize.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getPackingOptimizeNeighborUpdateInterval() << "\n";

						std::string datFileName = m_exploreSchedule.getResultDirectoryForDatFiles() + std::to_string(packingFraction) + ".dat";
						std::string cifFileName = m_exploreSchedule.getResultDirectoryForCifFiles() + std::to_string(packingFraction) + ".cif";

						MakeDatFile makeDatFile;
						makeDatFile(datFileName, packingStructure, optimizeMessage.str());

						MakeCifFile makeCifFile;
						makeCifFile(cifFileName, packingStructure);

						{
							std::stringstream message;
							message << "Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());
						}
					}
				}
			}
		}
	}

	// main search in the densest packings
// ***************************************************************************************************************************************************************************************

}


void PackingFactory::executeWithMd()
{
	std::string resultDirectoryForMdFiles = m_exploreSchedule.getResultDirectory();
	{
		std::string childDirectoryName = "MdFiles/";

		MakeDirectory makeDirectory(m_exploreSchedule.getResultDirectory());
		makeDirectory(childDirectoryName);

		resultDirectoryForMdFiles += childDirectoryName;
	}

// *****************************************************************************************************************************************************************************************
	// full optimization in beginning

	while (m_exploreSchedule.getInitialSupplyingNumber() < m_exploreSchedule.getInitialDemandNumber())
	{
		int supplyingLabel = 0;
		m_exploreSchedule.initialProduction(supplyingLabel);

		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> annealingDist(m_exploreVariables.getAnnealingMaxDisplacementLower(), m_exploreVariables.getAnnealingMaxDisplacementUpper());
		std::uniform_real_distribution<double> packingOptimizeDist(m_exploreVariables.getPackingOptimizeMaxDisplacementLower(), m_exploreVariables.getPackingOptimizeMaxDisplacementUpper());


		PilingUpGenerate pilingUpGenerate(m_basisComposition);
		PackingStructure packingStructure = pilingUpGenerate();

		MakeMdFile makeMdFile(packingStructure);

		double annealingMaxDisplacement = annealingDist(engine);
		Annealing annealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), annealingMaxDisplacement);

		if (!annealing(makeMdFile))
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double secondAnnealingMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize secondAnnealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), 0.9999, secondAnnealingMaxDisplacement);

		if (!secondAnnealing(makeMdFile))
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double initialPackingOptimizeMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize initialPackingOptimize(packingStructure, m_exploreVariables.getInitialPackingOptimizeIterationNumber(), m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getInitialPackingOptimizeDecreaseFactor(), initialPackingOptimizeMaxDisplacement);

		if (!initialPackingOptimize(makeMdFile))
		{
			m_calculationAnalysis.appendInitialPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		const double smallAnnealingPackingFraction = packingStructure.getPackingFraction();
		m_calculationAnalysis.appendInitialPackingOptimizePackingFractionsDistribution(smallAnnealingPackingFraction);

		double packingOptimizeMaxDisplacement = packingOptimizeDist(engine);
		IterativeBalanceOptimize spherePackingOptimize(packingStructure, m_exploreVariables.getPackingOptimizeIterationNumber(), m_exploreVariables.getPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getPackingOptimizeDecreaseFactor(), packingOptimizeMaxDisplacement);

		if (!spherePackingOptimize(makeMdFile))
		{
			m_calculationAnalysis.appendPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
				continue;
			}

			else
			{
				if (minimumZ < m_exploreSchedule.getOutputMinimumZ())
				{
					m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
					continue;
				}

				else
				{
					const double packingFraction = packingStructure.getPackingFraction();

					m_calculationAnalysis.appendPackingFractionsError(smallAnnealingPackingFraction, packingFraction);
					m_calculationAnalysis.appendPackingOptimizePackingFractionsDistribution(packingFraction);

					if ((packingFraction - m_exploreSchedule.getMaxPackingFraction()) < -1.0e-5)
					{
						continue;
					}

					else if (-1.0e-5 < (packingFraction - m_exploreSchedule.getMaxPackingFraction()) && (packingFraction - m_exploreSchedule.getMaxPackingFraction()) < 1.0e-5)
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);

						std::stringstream message;
						message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find the putative densest packing fraction again of " << packingFraction << "\n";
						m_reportCurrentSituation.report(message.str());

						continue;
					}

					else
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.normalizeDensePackingStructures();

						if (packingFraction < m_exploreSchedule.getOutputMinimumPackingFraction())
						{
							std::stringstream message;
							message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());

							continue;
						}

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << supplyingLabel;

						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
						optimizeMessage << "Annealing.Maximum.Displacement                                     " << annealingMaxDisplacement << "\n";
						optimizeMessage << "Annealing.Iteration.Number                                         " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "Annealing.Interval.To.Update.The.Neighboringlist                   " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "SecondAnnealing.Maximum.Displacement                               " << secondAnnealingMaxDisplacement << "\n";
						optimizeMessage << "SecondAnnealing.Iteration.Number                                   " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "SecondAnnealing.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "InitialPackingOptimize.Maximum.Displacement                        " << initialPackingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "InitialPackingOptimize.Decrease.Factor                             " << m_exploreVariables.getInitialPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "InitialPackingOptimize.Iteration.Number                            " << m_exploreVariables.getInitialPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "InitialPackingOptimize.Interval.To.Update.The.Neighboringlist      " << m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval() << "\n";

						optimizeMessage << "PackingOptimize.Maximum.Displacement                               " << packingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "PackingOptimize.Decrease.Factor                                    " << m_exploreVariables.getPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "PackingOptimize.Iteration.Number                                   " << m_exploreVariables.getPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "PackingOptimize.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getPackingOptimizeNeighborUpdateInterval() << "\n";


						std::string datFileName = m_exploreSchedule.getResultDirectoryForDatFiles() + std::to_string(packingFraction) + ".dat";
						std::string cifFileName = m_exploreSchedule.getResultDirectoryForCifFiles() + std::to_string(packingFraction) + ".cif";
						std::string mdFileName = resultDirectoryForMdFiles + std::to_string(packingFraction) + ".md";

						MakeDatFile makeDatFile;
						makeDatFile(datFileName, packingStructure, optimizeMessage.str());

						MakeCifFile makeCifFile;
						makeCifFile(cifFileName, packingStructure);

						makeMdFile(mdFileName);

						{
							std::stringstream message;
							message << "Initial Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());
						}
					}
				}
			}
		}
	}

	// full optimization in beginning
// ***************************************************************************************************************************************************************************************

// *****************************************************************************************************************************************************************************************
	// main search in the densest packings

	m_calculationAnalysis.setThresholdPackingFraction();

	while (m_exploreSchedule.getSupplyingNumber() < m_exploreSchedule.getDemandNumber())
	{
		int supplyingLabel = 0;
		m_exploreSchedule.production(supplyingLabel);

		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> annealingDist(m_exploreVariables.getAnnealingMaxDisplacementLower(), m_exploreVariables.getAnnealingMaxDisplacementUpper());
		std::uniform_real_distribution<double> packingOptimizeDist(m_exploreVariables.getPackingOptimizeMaxDisplacementLower(), m_exploreVariables.getPackingOptimizeMaxDisplacementUpper());


		PilingUpGenerate pilingUpGenerate(m_basisComposition);
		PackingStructure packingStructure = pilingUpGenerate();

		MakeMdFile makeMdFile(packingStructure);

		double annealingMaxDisplacement = annealingDist(engine);
		Annealing annealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), annealingMaxDisplacement);

		if (!annealing(makeMdFile))
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double secondAnnealingMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize secondAnnealing(packingStructure, m_exploreVariables.getAnnealingIterationNumber(), m_exploreVariables.getAnnealingNeighborUpdateInterval(), 0.9999, secondAnnealingMaxDisplacement);

		if (!secondAnnealing(makeMdFile))
		{
			m_calculationAnalysis.appendAnnealingMinusVolumeRejectNumber();
			continue;
		}

		double initialPackingOptimizeMaxDisplacement = annealingDist(engine);
		IterativeBalanceOptimize initialPackingOptimize(packingStructure, m_exploreVariables.getInitialPackingOptimizeIterationNumber(), m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getInitialPackingOptimizeDecreaseFactor(), initialPackingOptimizeMaxDisplacement);

		if (!initialPackingOptimize(makeMdFile))
		{
			m_calculationAnalysis.appendInitialPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		const double smallAnnealingPackingFraction = packingStructure.getPackingFraction();
		m_calculationAnalysis.appendInitialPackingOptimizePackingFractionsDistribution(smallAnnealingPackingFraction);

		if (smallAnnealingPackingFraction < m_calculationAnalysis.getThresholdPackingFraction())
		{
			m_calculationAnalysis.appendLowPackingFractionRejectNumber();
			continue;
		}

		double packingOptimizeMaxDisplacement = packingOptimizeDist(engine);
		IterativeBalanceOptimize spherePackingOptimize(packingStructure, m_exploreVariables.getPackingOptimizeIterationNumber(), m_exploreVariables.getPackingOptimizeNeighborUpdateInterval(), m_exploreVariables.getPackingOptimizeDecreaseFactor(), packingOptimizeMaxDisplacement);

		if (!spherePackingOptimize(makeMdFile))
		{
			m_calculationAnalysis.appendPackingOptimizeMinusVolumeRejectNumber();
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
				continue;
			}

			else
			{
				if (minimumZ < m_exploreSchedule.getOutputMinimumZ())
				{
					m_calculationAnalysis.appendPackingOptimizeZvalueRejectNumber();
					continue;
				}

				else
				{
					const double packingFraction = packingStructure.getPackingFraction();

					m_calculationAnalysis.appendPackingFractionsError(smallAnnealingPackingFraction, packingFraction);
					m_calculationAnalysis.appendPackingOptimizePackingFractionsDistribution(packingFraction);
					m_calculationAnalysis.setThresholdPackingFraction();

					if ((packingFraction - m_exploreSchedule.getMaxPackingFraction()) < -1.0e-5)
					{
						continue;
					}

					else if (-1.0e-5 < (packingFraction - m_exploreSchedule.getMaxPackingFraction()) && (packingFraction - m_exploreSchedule.getMaxPackingFraction()) < 1.0e-5)
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.appendMaxPackingFractionStructureNumber();

						std::stringstream message;
						message << "Trial Label  =  " << supplyingLabel << "  :  Find the putative densest packing fraction again of " << packingFraction << "\n";
						m_reportCurrentSituation.report(message.str());

						continue;
					}

					else
					{
						m_exploreSchedule.setDensePackingStructures(packingStructure);
						m_exploreSchedule.normalizeDensePackingStructures();
						m_exploreSchedule.demandReceive();

						m_exploreSchedule.renewMaxPackingFractionStructureNumber();

						if (packingFraction < m_exploreSchedule.getOutputMinimumPackingFraction())
						{
							std::stringstream message;
							message << "Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());

							continue;
						}

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << supplyingLabel;

						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";
						optimizeMessage << "Annealing.Maximum.Displacement                                     " << annealingMaxDisplacement << "\n";
						optimizeMessage << "Annealing.Iteration.Number                                         " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "Annealing.Interval.To.Update.The.Neighboringlist                   " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "SecondAnnealing.Maximum.Displacement                               " << secondAnnealingMaxDisplacement << "\n";
						optimizeMessage << "SecondAnnealing.Iteration.Number                                   " << m_exploreVariables.getAnnealingIterationNumber() << "\n";
						optimizeMessage << "SecondAnnealing.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "InitialPackingOptimize.Maximum.Displacement                        " << initialPackingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "InitialPackingOptimize.Decrease.Factor                             " << m_exploreVariables.getInitialPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "InitialPackingOptimize.Iteration.Number                            " << m_exploreVariables.getInitialPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "InitialPackingOptimize.Interval.To.Update.The.Neighboringlist      " << m_exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval() << "\n";

						optimizeMessage << "PackingOptimize.Maximum.Displacement                               " << packingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "PackingOptimize.Decrease.Factor                                    " << m_exploreVariables.getPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "PackingOptimize.Iteration.Number                                   " << m_exploreVariables.getPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "PackingOptimize.Interval.To.Update.The.Neighboringlist             " << m_exploreVariables.getPackingOptimizeNeighborUpdateInterval() << "\n";

						std::string datFileName = m_exploreSchedule.getResultDirectoryForDatFiles() + std::to_string(packingFraction) + ".dat";
						std::string cifFileName = m_exploreSchedule.getResultDirectoryForCifFiles() + std::to_string(packingFraction) + ".cif";
						std::string mdFileName = resultDirectoryForMdFiles + std::to_string(packingFraction) + ".md";

						MakeDatFile makeDatFile;
						makeDatFile(datFileName, packingStructure, optimizeMessage.str());

						MakeCifFile makeCifFile;
						makeCifFile(cifFileName, packingStructure);

						makeMdFile(mdFileName);

						{
							std::stringstream message;
							message << "Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction of " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());
						}
					}
				}
			}
		}
	}

	// main search in the densest packings
// ***************************************************************************************************************************************************************************************

}
