#define _USE_MATH_DEFINES
#include <cmath>
#include <memory>
#include <map>
#include <set>

#include "PackingStructure.h"

using namespace rk::dnspack;

PackingStructure::PackingStructure()
	: m_bravaisLattice()
	, m_basis()
{
}

PackingStructure::PackingStructure(const BravaisLattice& bravaisLattice, const std::vector<Sphere>& basis)
	: m_bravaisLattice(bravaisLattice)
	, m_basis(basis)
{
}

PackingStructure::PackingStructure(const BravaisLattice& bravaisLattice, std::vector<Sphere>&& basis) noexcept
	: m_bravaisLattice(bravaisLattice)
	, m_basis(std::move(basis))
{
}

PackingStructure::~PackingStructure()
{
}

BasisComposition PackingStructure::getBasisComposition() const
{
	std::map<double, int> composition;

	for (const auto& sphere : m_basis)
	{
		double radius = sphere.getRadius();

		auto iter = composition.find(radius);

		if (iter == composition.end())
		{
			composition.emplace(radius, 1);
		}

		else
		{
			iter->second += 1;
		}
	}

	return BasisComposition(std::move(composition));
}

double PackingStructure::getPackingFraction() const noexcept
{
	double fillingFactor = 0;

	for (auto& sphere : m_basis)
	{
		double radius = sphere.getRadius();
		fillingFactor += (4 * M_PI * radius * radius * radius) / 3;
	}

	fillingFactor /= m_bravaisLattice.getVolume();

	return fillingFactor;
}

bool PackingStructure::overlapExistence(double& overlapVolume, double& minimumZ) const noexcept
{
	overlapVolume = 0.0;
	minimumZ = 0.0;

	int spheresNumber = static_cast<int>(getSpheresNumber());

	int Q[3];
	{
		CartesianCoordinate reciprocalLatticeVector[3];
		{
			double volume = m_bravaisLattice.getVolume();

			reciprocalLatticeVector[0] = crossProduct(m_bravaisLattice[1], m_bravaisLattice[2]);
			reciprocalLatticeVector[0].scaling((1 / volume));

			reciprocalLatticeVector[1] = crossProduct(m_bravaisLattice[2], m_bravaisLattice[0]);
			reciprocalLatticeVector[1].scaling((1 / volume));

			reciprocalLatticeVector[2] = crossProduct(m_bravaisLattice[0], m_bravaisLattice[1]);
			reciprocalLatticeVector[2].scaling((1 / volume));
		}

		Q[0] = static_cast<int>(2.1 * reciprocalLatticeVector[0].getNorm()) + 1;
		Q[1] = static_cast<int>(2.1 * reciprocalLatticeVector[1].getNorm()) + 1;
		Q[2] = static_cast<int>(2.1 * reciprocalLatticeVector[2].getNorm()) + 1;
	}

	for (int baseLabel = 0; baseLabel < spheresNumber; ++baseLabel)
	{
		for (int Qa = -Q[0]; Qa <= Q[0]; ++Qa)
		{
			for (int Qb = -Q[1]; Qb <= Q[1]; ++Qb)
			{
				for (int Qc = -Q[2]; Qc <= Q[2]; ++Qc)
				{
					FractionalCoordinate Q{ static_cast<double>(Qa), static_cast<double>(Qb), static_cast<double>(Qc) };

					for (int sphereLabel = baseLabel + 1; sphereLabel < spheresNumber; ++sphereLabel)
					{
						FractionalCoordinate relativeFractionalVector(0, 0, 0);
						relativeFractionalVector = m_basis[sphereLabel].getFractionalCoordinate() + Q - m_basis[baseLabel].getFractionalCoordinate();

						CartesianCoordinate relativeVector(0, 0, 0);
						{
							relativeVector += scaling(relativeFractionalVector.a, m_bravaisLattice[0]);
							relativeVector += scaling(relativeFractionalVector.b, m_bravaisLattice[1]);
							relativeVector += scaling(relativeFractionalVector.c, m_bravaisLattice[2]);
						}

						double r1 = m_basis[baseLabel].getRadius();
						double r2 = m_basis[sphereLabel].getRadius();

						double distanceSquare = relativeVector.getNormSquare();
						double radiusSum = r1 + r2;

						if ((distanceSquare - radiusSum * radiusSum) < 0)
						{
							double distance = sqrt(distanceSquare);
							double zValue = distance - radiusSum;

							if (-0.0001 < zValue)
							{
								double z1 = ((distance * distance + r1 * r1 - r2 * r2) / (2 * distance));
								double z2 = ((distance * distance + r2 * r2 - r1 * r1) / (2 * distance));

								overlapVolume += std::abs((M_PI * (r1 - z1) * (r1 - z1) * (2 * r1 + z1)) / 3);
								overlapVolume += std::abs((M_PI * (r2 - z2) * (r2 - z2) * (2 * r2 + z2)) / 3);

								if (zValue < minimumZ)
								{
									minimumZ = zValue;
								}
							}

							else
							{
								return true;
							}
						}
					}

					if (!(Qa == 0 && Qb == 0 && Qc == 0))
					{
						CartesianCoordinate relativeVector(0, 0, 0);
						{
							relativeVector += scaling(Q.a, m_bravaisLattice[0]);
							relativeVector += scaling(Q.b, m_bravaisLattice[1]);
							relativeVector += scaling(Q.c, m_bravaisLattice[2]);
						}

						double r = m_basis[baseLabel].getRadius();

						double distanceSquare = relativeVector.getNormSquare();
						double radiusSum = 2.0 * r;

						if ((distanceSquare - radiusSum * radiusSum) < 0)
						{
							double distance = sqrt(distanceSquare);
							double zValue = distance - radiusSum;

							if (-0.0001 < zValue)
							{
								double z = 0.5 * distance;
								overlapVolume += std::abs((M_PI * (r - z) * (r - z) * (2 * r + z)) / 3);

								if (zValue < minimumZ)
								{
									minimumZ = zValue;
								}
							}

							else
							{
								return true;
							}
						}
					}
				}
			}
		}
	}

	return false;
}
