#define _USE_MATH_DEFINES

#include "PilingUpGenerate.h"

#include "EnvironmentVariables.h"

#include <map>
#include <vector>
#include <utility>
#include <random>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <cmath>

using namespace rk::dnspack;


PilingUpGenerate::PilingUpGenerate(const BasisComposition& basisComposition)
	: m_basisComposition(basisComposition)
	, m_spheresNumber(basisComposition.getSpheresNumber())
{
	EnvironmentVariables environmentVariables;
	m_uniformGenerationRatio = environmentVariables.getExploreVariables().getUniformGenerationRatio();
}

PilingUpGenerate::~PilingUpGenerate()
{
}

PackingStructure PilingUpGenerate::operator()()
{
	std::random_device seed_gen;
	std::default_random_engine engine(seed_gen());

	std::uniform_real_distribution<double> coordinateDist(0.0, 1.0);

// ********************************************************************************************************************************************************************
	// initialization

	BravaisLattice bravaisLattice;

	std::vector<Sphere> basis;
	basis.reserve(m_spheresNumber);
	int nonPlacedSpheresNumber = static_cast<int>(m_spheresNumber);

	std::vector<std::pair<int, double>> layerInfo;
	int firstLayerSpheresNumber = 0;
	double maxArea = 0.0;

	std::vector<double> radii;

	for (const auto& radiiAndNum : m_basisComposition.getComposition())
	{
		double radius = radiiAndNum.first;
		int number = radiiAndNum.second;

		for (int num = 0; num < number; ++num)
		{
			radii.push_back(radius);
		}
	}

	std::shuffle(radii.begin(), radii.end(), engine);

	// initialization
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
	// Generation of the first layer

	while (firstLayerSpheresNumber == 0)
	{
		{
			std::binomial_distribution<> generateTypeDist(1, m_uniformGenerationRatio);

			if (0 == generateTypeDist(engine))
			{
				std::binomial_distribution<> dist(static_cast<int>(m_spheresNumber), 0.5);
				firstLayerSpheresNumber = dist(engine);

				if (firstLayerSpheresNumber == 0)
				{
					continue;
				}
			}

			else
			{
				std::uniform_int_distribution<> dist(1, static_cast<int>(m_spheresNumber));
				firstLayerSpheresNumber = dist(engine);
			}
		}

		double maxRadius = 0.0;
		double area = 0.0;

		for (int label = 0; label < firstLayerSpheresNumber; ++label)
		{
			double radius = radii.back();
			{
				if (maxRadius < radius)
				{
					maxRadius = radius;
				}

				area += (2.0 * radius * radius);
			}

			Sphere sphere(radius);

			sphere.getFractionalCoordinate().a = coordinateDist(engine);
			sphere.getFractionalCoordinate().b = coordinateDist(engine);

			basis.push_back(sphere);

			radii.pop_back();
		}

		maxArea = area;
		layerInfo.push_back(std::make_pair(firstLayerSpheresNumber, maxRadius));
		nonPlacedSpheresNumber -= firstLayerSpheresNumber;
	}

	// Generation of the first layer
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
	// Generation of the other layers

	while (0 < nonPlacedSpheresNumber)
	{
		int placedSpheresNumber = 0;
		{
			if (nonPlacedSpheresNumber == 1)
			{
				placedSpheresNumber = 1;
			}

			else
			{
				int maxPlacedNumber = std::min(nonPlacedSpheresNumber, firstLayerSpheresNumber);

				std::uniform_int_distribution<> dist(1, maxPlacedNumber);
				placedSpheresNumber = dist(engine);
			}
		}

		if (placedSpheresNumber == 0)
		{
			continue;
		}

		else
		{
			double maxRadius = 0.0;
			double area = 0.0;

			for (int label = 0; label < placedSpheresNumber; ++label)
			{
				double radius = radii.back();
				{
					if (maxRadius < radius)
					{
						maxRadius = radius;
					}

					area += (2.0 * radius * radius);
				}

				Sphere sphere(radius);

				sphere.getFractionalCoordinate().a = coordinateDist(engine);
				sphere.getFractionalCoordinate().b = coordinateDist(engine);

				basis.push_back(sphere);

				radii.pop_back();
			}


			if (maxArea < area)
			{
				maxArea = area;
			}

			layerInfo.push_back(std::make_pair(placedSpheresNumber, maxRadius));
			nonPlacedSpheresNumber -= placedSpheresNumber;
		}
	}

	// Generation of the other layers
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
	// Decision of xy components of lattice vectors

	{
		double minAngle = (M_PI / 3.0);
		std::uniform_real_distribution<double> angleDist(minAngle, (2.0 * minAngle));
		
		double length = sqrt(maxArea);
		double angle = angleDist(engine);

		bravaisLattice[0].x = length;
		bravaisLattice[1].x = length * std::cos(angle);
		bravaisLattice[1].y = length * std::sin(angle);
	}

	// Decision of xy components of lattice vectors
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
	// Decision of z component of lattice vector

	double maxRadiusSum = 0.0;
	{
		for (const auto& numAndMaxRadius : layerInfo)
		{
			maxRadiusSum += numAndMaxRadius.second;
		}
	}

	bravaisLattice[2].z = maxRadiusSum;

	std::uniform_real_distribution<double> tiltingDist(-0.50, 0.50);
	{
		bravaisLattice[2] += scaling(tiltingDist(engine), bravaisLattice[0]);
		bravaisLattice[2] += scaling(tiltingDist(engine), bravaisLattice[1]);
	}

	// Decision of z component of lattice vector
// ********************************************************************************************************************************************************************


// ********************************************************************************************************************************************************************
	// Decision of z fractional coordinates

	double twiceMaxRadiusSum = 2.0 * maxRadiusSum;
	double zFractionalCoordinate = 0.0;
	{
		zFractionalCoordinate -= (layerInfo.front().second / twiceMaxRadiusSum);
		zFractionalCoordinate += 0.0001;
	}

	std::vector<Sphere>::iterator basisIter = basis.begin();

	for (const auto& numAndMaxRadius : layerInfo)
	{
		int num = numAndMaxRadius.first;
		double maxRadius = numAndMaxRadius.second;

		zFractionalCoordinate += (maxRadius / twiceMaxRadiusSum);

		for (int label = 0; label < num; ++label)
		{
			basisIter->getFractionalCoordinate().c = zFractionalCoordinate;
			++basisIter;
		}

		zFractionalCoordinate += (maxRadius / twiceMaxRadiusSum);
	}

	// Decision of z fractional coordinates
// ********************************************************************************************************************************************************************


	return PackingStructure(bravaisLattice, std::move(basis));
}
