#include "ReOptimizeFactory.h"

#include <sstream>
#include <random>
#include <thread>

#include "RandomFluctuate.h"
#include "Annealing.h"
#include "IterativeBalanceOptimize.h"

#include "MakeDatFile.h"
#include "MakeCifFile.h"

using namespace rk::dnspack;


ReOptimizeFactory::ReOptimizeFactory(const ReOptimizeSchedule& reOptimizeSchedule)
	: m_reOptimizeSchedule(reOptimizeSchedule)
	, m_reportCurrentSituation(reOptimizeSchedule.getDensestPackingsDirectory() + "currentSituation.txt")
{
	EnvironmentVariables environmentVariables;
	m_reOptimizeVariables = environmentVariables.getReOptimizeVariables();
}

ReOptimizeFactory::~ReOptimizeFactory()
{
}


void ReOptimizeFactory::operator()()
{
	const std::thread::id myThreadID{ std::this_thread::get_id() };

	{
		std::stringstream reportMessage;
		reportMessage << "Thread ID  =  " << myThreadID << "  :  Begin to re-optimize packing structures. \n";
		m_reportCurrentSituation.report(reportMessage.str());
	}

// ******************************************************************************************************************************************
	// re-optimization

	RandomFluctuate randomFluctuate(m_reOptimizeVariables.getMaxFluctuation());

	while (m_reOptimizeSchedule.getSupplyingNumber() < m_reOptimizeSchedule.getDemandNumber())
	{
		int supplyingLabel = 0;
		m_reOptimizeSchedule.reOptimization(supplyingLabel);

		std::random_device seed_gen;
		std::default_random_engine engine(seed_gen());

		std::uniform_real_distribution<double> initialAnnealingDist(m_reOptimizeVariables.getInitialAnnealingMaxDisplacementLower(), m_reOptimizeVariables.getInitialAnnealingMaxDisplacementUpper());
		std::uniform_real_distribution<double> densestPackingOptimizeDist(m_reOptimizeVariables.getDensestPackingOptimizeMaxDisplacementLower(), m_reOptimizeVariables.getDensestPackingOptimizeMaxDisplacementUpper());


		PackingStructure packingStructure = m_reOptimizeSchedule.getSeedStructure();
		randomFluctuate(packingStructure);

		double initialAnnealingMaxDisplacement = initialAnnealingDist(engine);
		Annealing initialAnnealing(packingStructure, m_reOptimizeVariables.getInitialAnnealingIterationNumber(), m_reOptimizeVariables.getInitialAnnealingNeighborUpdateInterval(), initialAnnealingMaxDisplacement);

		if (!initialAnnealing())
		{
			continue;
		}

		double densestPackingOptimizeMaxDisplacement = densestPackingOptimizeDist(engine);
		IterativeBalanceOptimize densestPackingOptimize(packingStructure, m_reOptimizeVariables.getDensestPackingOptimizeIterationNumber(), m_reOptimizeVariables.getDensestPackingOptimizeNeighborUpdateInterval(), m_reOptimizeVariables.getDensestPackingOptimizeDecreaseFactor(), densestPackingOptimizeMaxDisplacement);

		if (!densestPackingOptimize())
		{
			continue;
		}

		else
		{
			double overlapVolume = 0.0;
			double minimumZ = 0.0;

			if (packingStructure.overlapExistence(overlapVolume, minimumZ))
			{
				continue;
			}

			else
			{
				if (minimumZ < m_reOptimizeSchedule.getOutputMinimumZ())
				{
					continue;
				}

				else
				{
					double packingFraction = packingStructure.getPackingFraction();

					if (m_reOptimizeSchedule.getMaxPackingFraction() < packingFraction)
					{
						m_reOptimizeSchedule.setMaxPackingFraction(packingFraction);

						std::stringstream optimizeMessage;

						optimizeMessage << "\n\n#\n" << "# Creation Information\n" << "#\n\n\n";

						optimizeMessage << "Trial.Label                                " << supplyingLabel;

						optimizeMessage << "\n\n#\n" << "# Optimize Paramaters\n" << "#\n\n\n";

						optimizeMessage << "InitialAnnealing.Maximum.Displacement                              " << initialAnnealingMaxDisplacement << "\n";
						optimizeMessage << "InitialAnnealing.Iteration.Number                                  " << m_reOptimizeVariables.getInitialAnnealingIterationNumber() << "\n";
						optimizeMessage << "InitialAnnealing.Interval.To.Update.The.Neighboringlist            " << m_reOptimizeVariables.getInitialAnnealingNeighborUpdateInterval() << "\n";

						optimizeMessage << "DensestPackingOptimize.Maximum.Displacement                        " << densestPackingOptimizeMaxDisplacement << "\n";
						optimizeMessage << "DensestPackingOptimize.Decrease.Factor                             " << m_reOptimizeVariables.getDensestPackingOptimizeDecreaseFactor() << "\n";
						optimizeMessage << "DensestPackingOptimize.Iteration.Number                            " << m_reOptimizeVariables.getDensestPackingOptimizeIterationNumber() << "\n";
						optimizeMessage << "DensestPackingOptimize.Interval.To.Update.The.Neighboringlist      " << m_reOptimizeVariables.getDensestPackingOptimizeNeighborUpdateInterval() << "\n";


						std::string datFileName = m_reOptimizeSchedule.getDensestPackingsDirectoryForDatFiles() + std::to_string(packingFraction) + ".dat";
						std::string cifFileName = m_reOptimizeSchedule.getDensestPackingsDirectoryForCifFiles() + std::to_string(packingFraction) + ".cif";

						MakeDatFile makeDatFile;
						makeDatFile(datFileName, packingStructure, optimizeMessage.str());

						MakeCifFile makeCifFile;
						makeCifFile(cifFileName, packingStructure);

						{
							std::stringstream message;
							message << "Trial Label  =  " << supplyingLabel << "  :  Find a denser packing fraction   " << packingFraction << "\n";
							m_reportCurrentSituation.report(message.str());
						}
					}
				}
			}
		}
	}

	// re-optimization
// ******************************************************************************************************************************************

}
