#include "ReadInputFile.h"

#include <fstream>
#include <sstream>
#include <regex>
#include <map>
#include <thread>
#include <stdexcept>
#include <algorithm>
#include <cmath>

using namespace rk::dnspack;

ReadInputFile::ReadInputFile(const std::string& fileName)
	: m_fileName{ fileName }
	, m_threadNumber{ 0 }
	, m_outputDirectoryName{}
	, m_exploreVariables()
	, m_reOptimizeVariables()
	, m_trialBasisCompositions()
{
}

ReadInputFile::~ReadInputFile()
{
}

void ReadInputFile::initialize()
{
// ******************************************************************************************************************************************
	// Read input file

	std::string fileString;
	{
		std::ifstream inputFileStream;
		inputFileStream.open(m_fileName, std::ios_base::in);
		if (!inputFileStream)
		{
			throw std::runtime_error("could not open the input file.");
		}

		for (std::string stringLine; std::getline(inputFileStream, stringLine);)
		{
			fileString += stringLine + "\n";
		}
		inputFileStream.close();
	}

	// Read input file
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Thread number

	{
		std::regex pat("[\\n][\\s]*Thread.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		m_threadNumber = static_cast<int>(std::thread::hardware_concurrency());

		if (std::regex_search(fileString, numSm, pat))
		{
			m_threadNumber = std::stoi(numSm.str(1));
		}

		else
		{
			if (m_threadNumber < 1)
			{
				throw std::runtime_error("Thread.Number is less than 1.");
			}
		}
	}

	// Thread number
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Directory path for output files

	{
		std::regex pat("[\\n][\\s]*Output.Directory.Name[\\s]+(.+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_outputDirectoryName = numSm.str(1);

			if (m_outputDirectoryName.back() != '/')
			{
				throw std::invalid_argument("The last literature of Output.Directory.Name is not /");
			}
		}

		else
		{
			throw std::runtime_error("couldn't get the output directory name.");
		}
	}

	// Directory path for output files
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Optimize parameters for exhaustive exploration

	{
		double annealingMaxDisplacementUpper = 0.10;
		{
			std::regex pat("[\\n][\\s]*Annealing.Maximum.Displacement.Upper[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				annealingMaxDisplacementUpper = std::stod(numSm.str(1));
			}
		}

		double annealingMaxDisplacementLower = 0.03;
		{
			std::regex pat("[\\n][\\s]*Annealing.Maximum.Displacement.Lower[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				annealingMaxDisplacementLower = std::stod(numSm.str(1));
			}
		}

		m_exploreVariables.setAnnealingMaxDisplacement(annealingMaxDisplacementUpper, annealingMaxDisplacementLower);
	}

	{
		std::regex pat("[\\n][\\s]*Annealing.Iteration.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setAnnealingIterationNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setAnnealingIterationNumber(500);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Annealing.Interval.To.Update.The.Neighboringlist[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setAnnealingNeighborUpdateInterval(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setAnnealingNeighborUpdateInterval(200);
		}
	}

	{
		std::regex pat("[\\n][\\s]*InitialPackingOptimize.Decrease.Factor[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setInitialPackingOptimizeDecreaseFactor(std::stod(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setInitialPackingOptimizeDecreaseFactor(0.9995);
		}
	}

	{
		std::regex pat("[\\n][\\s]*InitialPackingOptimize.Iteration.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setInitialPackingOptimizeIterationNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setInitialPackingOptimizeIterationNumber(2500);
		}
	}

	{
		std::regex pat("[\\n][\\s]*InitialPackingOptimize.Interval.To.Update.The.Neighboringlist[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setInitialPackingOptimizeNeighborUpdateInterval(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setInitialPackingOptimizeNeighborUpdateInterval(300);
		}
	}

	{
		double packingOptimizeMaxDisplacementUpper = 0.05;
		{
			std::regex pat("[\\n][\\s]*PackingOptimize.Maximum.Displacement.Upper[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				packingOptimizeMaxDisplacementUpper = std::stod(numSm.str(1));
			}
		}

		double packingOptimizeMaxDisplacementLower = 0.02;
		{
			std::regex pat("[\\n][\\s]*PackingOptimize.Maximum.Displacement.Lower[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				packingOptimizeMaxDisplacementLower = std::stod(numSm.str(1));
			}
		}

		m_exploreVariables.setPackingOptimizeMaxDisplacement(packingOptimizeMaxDisplacementUpper, packingOptimizeMaxDisplacementLower);
	}

	{
		std::regex pat("[\\n][\\s]*PackingOptimize.Decrease.Factor[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setPackingOptimizeDecreaseFactor(std::stod(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setPackingOptimizeDecreaseFactor(0.9997);
		}
	}

	{
		std::regex pat("[\\n][\\s]*PackingOptimize.Iteration.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setPackingOptimizeIterationNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setPackingOptimizeIterationNumber(40000);
		}
	}

	{
		std::regex pat("[\\n][\\s]*PackingOptimize.Interval.To.Update.The.Neighboringlist[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_exploreVariables.setPackingOptimizeNeighborUpdateInterval(std::stoi(numSm.str(1)));
		}

		else
		{
			m_exploreVariables.setPackingOptimizeNeighborUpdateInterval(300);
		}
	}

	// Optmize parameters for exhaustive exploration
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Conditions for exhaustive exploration

	{
		std::regex pat("[\\n][\\s]*Uniform.Generation.Ratio[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double uniformGenerationRatio = std::stod(numSm.str(1));

			if (1.0 < uniformGenerationRatio)
			{
				throw std::logic_error("fullOptimizeStructureNumber is more than one.");
			}

			else if (uniformGenerationRatio < 0.0)
			{
				throw std::logic_error("fullOptimizeStructureNumber is less than zero.");
			}

			else
			{
				m_exploreVariables.setUniformGenerationRatio(uniformGenerationRatio);
			}
		}

		else
		{
			throw std::runtime_error("couldn't get the Uniform.Generation.Ratio.");
		}
	}

	{
		std::regex pat("[\\n][\\s]*FullOptimize.Structure.Number.In.Beginning[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			int fullOptimizeStructureNumber = std::stoi(numSm.str(1));

			if (fullOptimizeStructureNumber < 1)
			{
				throw std::runtime_error("FullOptimize.Structure.Number.In.Beginning is less than one.");
			}

			m_exploreVariables.setFullOptimizeStructureNumber(fullOptimizeStructureNumber);
		}

		else
		{
			throw std::runtime_error("couldn't get the FullOptimize.Structure.Number.In.Beginning.");
		}
	}

	{
		std::regex pat("[\\n][\\s]*Maximum.Trial.Number.For.Each.Composition[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			int maxTrialNumber = std::stoi(numSm.str(1));

			if (maxTrialNumber < 1)
			{
				throw std::runtime_error("Maximum.Trial.Number.For.Each.Composition is less than one.");
			}

			m_exploreVariables.setMaxTrialNumber(maxTrialNumber);
		}

		else
		{
			throw std::runtime_error("couldn't get the Maximum.Trial.Number.For.Each.Composition.");
		}
	}

	{
		std::regex pat("[\\n][\\s]*Threshold.Trial.Number.For.Judging.Convergence[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			int thresholdTrialNumber = std::stoi(numSm.str(1));

			if (thresholdTrialNumber < 1)
			{
				throw std::runtime_error("Threshold.Trial.Number.For.Judging.Convergence is less than one.");
			}

			m_exploreVariables.setThresholdTrialNumber(thresholdTrialNumber);
		}

		else
		{
			throw std::runtime_error("couldn't get the Threshold.Trial.Number.For.Judging.Convergence.");
		}
	}

	{
		std::regex pat("[\\n][\\s]*Mesh.For.Packing.Fractions.Error[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double m_packingFractionsErrorMesh = std::stod(numSm.str(1));

			if (m_packingFractionsErrorMesh < 0.0)
			{
				throw std::runtime_error("Mesh.For.Packing.Fractions.Error is less than 0.0.");
			}

			m_exploreVariables.setPackingFractionsErrorMesh(m_packingFractionsErrorMesh);
		}

		else
		{
			m_exploreVariables.setPackingFractionsErrorMesh(0.01);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Mesh.For.InitialPackingOptimize.Packing.Fractions[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double initialPackingOptimizePackingFractionsMesh = std::stod(numSm.str(1));

			if (initialPackingOptimizePackingFractionsMesh < 0.0)
			{
				throw std::runtime_error("Mesh.For.InitialPackingOptimize.Packing.Fractions is less than 0.0.");
			}

			m_exploreVariables.setInitialPackingOptimizePackingFractionsMesh(initialPackingOptimizePackingFractionsMesh);
		}

		else
		{
			m_exploreVariables.setInitialPackingOptimizePackingFractionsMesh(0.01);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Mesh.For.PackingOptimize.Packing.Fractions[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double packingOptimizePackingFractionsMesh = std::stod(numSm.str(1));

			if (packingOptimizePackingFractionsMesh < 0.0)
			{
				throw std::runtime_error("Mesh.For.PackingOptimize.Packing.Fractions is less than 0.0.");
			}

			m_exploreVariables.setPackingOptimizePackingFractionsMesh(packingOptimizePackingFractionsMesh);
		}

		else
		{
			m_exploreVariables.setPackingOptimizePackingFractionsMesh(0.01);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Output.Minimum.Packing.Fraction[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double outputMinimumPackingFraction = std::stod(numSm.str(1));

			if (outputMinimumPackingFraction < 0.0)
			{
				throw std::runtime_error("Output.Minimum.Packing.Fraction is less than 0.0.");
			}

			m_exploreVariables.setOutputMinimumPackingFraction(outputMinimumPackingFraction);
		}

		else
		{
			m_exploreVariables.setOutputMinimumPackingFraction(0.73);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Output.MD.File[\\s]+([\\w]+)[\\s]*");
		std::smatch boolSm;

		if (std::regex_search(fileString, boolSm, pat))
		{
			std::string outputOption = boolSm.str(1);

			if (outputOption == "ON")
			{
				m_exploreVariables.setOutputMdFile(true);
			}

			else
			{
				if (outputOption == "OFF")
				{
					m_exploreVariables.setOutputMdFile(false);
				}

				else
				{
					throw std::runtime_error("could not read the option of Output.MD.File.");
				}
			}
		}

		else
		{
			throw std::runtime_error("could not read the option of Output.MD.File.");
		}
	}

	{
		std::regex pat("[\\n][\\s]*Output.Minimum.Overlap[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double outputMinimumOverlap = std::stod(numSm.str(1));

			if (outputMinimumOverlap <= 0.0)
			{
				throw std::runtime_error("Output.Minimum.Overlap is less than or equal to 0.0.");
			}

			outputMinimumOverlap *= (-1.0);
			m_exploreVariables.setOutputMinimumZ(outputMinimumOverlap);
		}

		else
		{
			m_exploreVariables.setOutputMinimumPackingFraction(-0.000005);
		}
	}


	// Conditions for exhaustive exploration
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Parameters for reoptimization

	{
		std::regex pat("[\\n][\\s]*Maximum.Fluctuation[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setMaxFluctuation(std::stod(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setMaxFluctuation(0.01);
		}
	}

	{
		double initialAnnealingMaxDisplacementUpper = 0.05;
		{
			std::regex pat("[\\n][\\s]*InitialAnnealing.Maximum.Displacement.Upper[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				initialAnnealingMaxDisplacementUpper = std::stod(numSm.str(1));
			}
		}

		double initialAnnealingMaxDisplacementLower = 0.003;
		{
			std::regex pat("[\\n][\\s]*InitialAnnealing.Maximum.Displacement.Lower[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				initialAnnealingMaxDisplacementLower = std::stod(numSm.str(1));
			}
		}

		m_reOptimizeVariables.setInitialAnnealingMaxDisplacement(initialAnnealingMaxDisplacementUpper, initialAnnealingMaxDisplacementLower);
	}

	{
		std::regex pat("[\\n][\\s]*InitialAnnealing.Iteration.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setInitialAnnealingIterationNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setInitialAnnealingIterationNumber(500);
		}
	}

	{
		std::regex pat("[\\n][\\s]*InitialAnnealing.Interval.To.Update.The.Neighboringlist[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setInitialAnnealingNeighborUpdateInterval(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setInitialAnnealingNeighborUpdateInterval(200);
		}
	}

	{
		double densestPackingOptimizeMaxDisplacementUpper = 0.02;
		{
			std::regex pat("[\\n][\\s]*DensestPackingOptimize.Maximum.Displacement.Upper[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				densestPackingOptimizeMaxDisplacementUpper = std::stod(numSm.str(1));
			}
		}

		double densestPackingOptimizeMaxDisplacementLower = 0.001;
		{
			std::regex pat("[\\n][\\s]*DensestPackingOptimize.Maximum.Displacement.Lower[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::smatch numSm;

			if (std::regex_search(fileString, numSm, pat))
			{
				densestPackingOptimizeMaxDisplacementLower = std::stod(numSm.str(1));
			}
		}

		m_reOptimizeVariables.setDensestPackingOptimizeMaxDisplacement(densestPackingOptimizeMaxDisplacementUpper, densestPackingOptimizeMaxDisplacementLower);
	}

	{
		std::regex pat("[\\n][\\s]*DensestPackingOptimize.Decrease.Factor[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setDensestPackingOptimizeDecreaseFactor(std::stod(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setDensestPackingOptimizeDecreaseFactor(0.9997);
		}
	}

	{
		std::regex pat("[\\n][\\s]*DensestPackingOptimize.Iteration.Number[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setDensestPackingOptimizeIterationNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setDensestPackingOptimizeIterationNumber(80000);
		}
	}

	{
		std::regex pat("[\\n][\\s]*DensestPackingOptimize.Interval.To.Update.The.Neighboringlist[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setDensestPackingOptimizeNeighborUpdateInterval(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setDensestPackingOptimizeNeighborUpdateInterval(1000);
		}
	}

	// Parameters for reoptimization
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Conditions for reoptimization

	{
		std::regex pat("[\\n][\\s]*Number.Of.ReOptimize.Structures[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setReOptimizeStructureNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setReOptimizeStructureNumber(20);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Number.Of.Re-Optimization.For.Each.Structure[\\s]+([\\d]+)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			m_reOptimizeVariables.setReOptimizeRepetitionNumber(std::stoi(numSm.str(1)));
		}

		else
		{
			m_reOptimizeVariables.setReOptimizeRepetitionNumber(3000);
		}
	}

	{
		std::regex pat("[\\n][\\s]*Output.Minimum.Overlap.For.ReOptimization[\\s]+([\\d]+[\\.]?[\\d]*)[\\s]*");
		std::smatch numSm;

		if (std::regex_search(fileString, numSm, pat))
		{
			double outputMinimumOverlap = std::stod(numSm.str(1));

			if (outputMinimumOverlap < 0.0)
			{
				throw std::runtime_error("Output.Minimum.Overlap is less than or equal to 0.0.");
			}

			outputMinimumOverlap *= (-1.0);
			m_reOptimizeVariables.setOutputMinimumZ(outputMinimumOverlap);
		}

		else
		{
			m_reOptimizeVariables.setOutputMinimumZ(-0.0000000001);
		}
	}

	// Conditions for reoptimization
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Compositions
	
	int sphereKindNumber = 0;
	std::vector<std::vector<int>> spheresNumbers;

	{
		std::size_t initialPos = fileString.find("<The.Number.Of.Spheres.In.Unit.Cell");
		std::size_t endPos = fileString.find("The.Number.Of.Spheres.In.Unit.Cell>");

		if (endPos <= initialPos)
		{
			throw std::runtime_error("couldn't get the number of spheres in an unit cell.");
		}

		if (initialPos != std::string::npos && endPos != std::string::npos)
		{
			std::size_t tempPos = fileString.find_first_of("\n", initialPos);
			std::string compositionString{ fileString, (tempPos + 1), (endPos - tempPos - 1) };
			std::stringstream compositionStream{ compositionString };

			std::regex totalPat("[\\s]*([\\d]+[\\s]*)*");
			std::regex iterPat("([\\d]+)");

			{
				std::string line;
				std::getline(compositionStream, line);

				std::smatch sm;

				if (std::regex_match(line, sm, totalPat))
				{
					std::vector<int> spheresNumber;

					for (std::sregex_iterator iter(line.begin(), line.end(), iterPat); iter != std::sregex_iterator{}; ++iter)
					{
						spheresNumber.push_back(std::stoi(iter->str(1)));
					}

					sphereKindNumber = static_cast<int>(spheresNumber.size());

					if (sphereKindNumber < 1)
					{
						throw std::runtime_error("couldn't get the number of spheres in an unit cell.");
					}

					else
					{
						if (6 < sphereKindNumber)
						{
							throw std::runtime_error("The kind of spheres is more than 6.");
						}

						spheresNumbers.push_back(spheresNumber);
					}
				}

				else
				{
					throw std::runtime_error("couldn't get the number of spheres in an unit cell.");
				}
			}

			for (std::string line; std::getline(compositionStream, line);)
			{
				std::smatch sm;

				if (std::regex_match(line, sm, totalPat))
				{
					std::vector<int> spheresNumber;

					for (std::sregex_iterator iter(line.begin(), line.end(), iterPat); iter != std::sregex_iterator{}; ++iter)
					{
						spheresNumber.push_back(std::stoi(iter->str(1)));
					}

					if (sphereKindNumber != static_cast<int>(spheresNumber.size()))
					{
						throw std::runtime_error("couldn't get the number of spheres in an unit cell.");
					}

					spheresNumbers.push_back(spheresNumber);
				}

				else
				{
					throw std::runtime_error("couldn't get the number of spheres in an unit cell.");
				}
			}
		}
	}

	// Compositions
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Radii ratios

	std::vector<std::vector<double>> radiiRatios;

	{
		std::size_t initialPos = fileString.find("<Radii.Ratios");
		std::size_t endPos = fileString.find("Radii.Ratios>");

		if (endPos <= initialPos)
		{
			throw std::runtime_error("couldn't get the radii ratios.");
		}

		if (initialPos != std::string::npos && endPos != std::string::npos)
		{
			std::size_t tempPos = fileString.find_first_of("\n", initialPos);
			std::string radiiString{ fileString, (tempPos + 1), (endPos - tempPos - 1) };
			std::stringstream radiiStream{ radiiString };

			std::regex totalPat("[\\s]*([\\d]+[\\.]?[\\d]*[\\s]+)*([\\d]+[\\.]?[\\d]*)[\\s]*");
			std::regex iterPat("([\\d]+[\\.]?[\\d]*)");

			for (std::string line; std::getline(radiiStream, line);)
			{
				std::smatch radiiSm;

				if (std::regex_match(line, radiiSm, totalPat))
				{
					std::vector<double> radiiRatio;

					for (std::sregex_iterator iter(line.begin(), line.end(), iterPat); iter != std::sregex_iterator{}; ++iter)
					{
						radiiRatio.push_back(std::stod(iter->str(1)));
					}

					if (sphereKindNumber != static_cast<int>(radiiRatio.size()))
					{
						throw std::runtime_error("couldn't get the radii ratios.");
					}

					else
					{
						double maximumRadiusDifference = radiiRatio.back() - 1.00;

						if (std::abs(maximumRadiusDifference) > 1.0e-14)
						{
							throw std::runtime_error("The maximum radius is not 1.");
						}

						else
						{
							if (!std::is_sorted(radiiRatio.begin(), radiiRatio.end()))
							{
								throw std::runtime_error("The radii is not sorted.");
							}

							else
							{
								std::vector<double>::iterator iter = radiiRatio.begin();
								std::vector<double>::iterator iterEnd = radiiRatio.end();

								for (; iter != std::prev(iterEnd); ++iter)
								{
									if (*iter == *(std::next(iter)))
									{
										throw std::runtime_error("The radii in each row have to be different each other.");
									}
								}

								radiiRatios.push_back(radiiRatio);
							}
						}
					}
				}

				else
				{
					throw std::runtime_error("couldn't get the radii ratios.");
				}
			}
		}

		else
		{
			throw std::runtime_error("couldn't get the radii ratios.");
		}
	}

	// Radii ratios
// ******************************************************************************************************************************************

// ******************************************************************************************************************************************
	// Set trialbasisCompositions

	m_trialBasisCompositions.clear();

	for (const auto& radiiRatio : radiiRatios)
	{
		for (const auto& sphereNumber : spheresNumbers)
		{
			int smallerSpheresNumber = 0;
			int largestSpheresNumber = sphereNumber.back();

			std::vector<int>::const_iterator iterEnd = sphereNumber.cend();


			for (std::vector<int>::const_iterator iter = sphereNumber.cbegin(); iter != std::prev(iterEnd); ++iter)
			{
				smallerSpheresNumber += *iter;
			}

			if (smallerSpheresNumber < largestSpheresNumber)
			{
				continue;
			}

			std::map<double, int> composition;

			std::vector<int>::const_iterator iter = sphereNumber.cbegin();

			for (const auto& radius : radiiRatio)
			{
				composition.emplace(radius, *iter);
				++iter;
			}

			m_trialBasisCompositions.push_back(BasisComposition(std::move(composition)));
		}
	}

	// Set trialbasisCompositions
// ******************************************************************************************************************************************

}
