#include "SAMLAIdsp.h"

#include <stdexcept>
#include <iostream>

#include "ReadInputFile.h"
#include "EnvironmentVariables.h"

#include "SAMLAIReadme.h"
#include "CalculationAnalysis.h"
#include "TaskManager.h"

#include "Clock.h"
#include "MakeDirectory.h"
#include "IsDirectory.h"


using namespace rk::dnspack;

SAMLAIdsp::SAMLAIdsp(rk::dnspack::IArgument& argument)
	: m_argument(argument)
{
}

SAMLAIdsp::~SAMLAIdsp()
{
}

int SAMLAIdsp::execute()
{
// ********************************************************************************************************************************************************************
	// Initial Message

	{
		std::string message;

		message += "*********************************************************************************************************\n\n";
		message += "This explore of the densest sphere packings was performed by SAMLAIsp " + SAMLAIVersion + "\n";
		message += Clock::getNowTime() + "\n\n";
		message += "*********************************************************************************************************\n\n\n";

		std::cout << message;
	}

	// Initial Message
// ********************************************************************************************************************************************************************

	EnvironmentVariables environmentVariables;
	std::vector<BasisComposition> trialBasisCompositions;

// ********************************************************************************************************************************************************************
	// Read input file

	{
		ReadInputFile readInputFile(m_argument.getInputFileName());
		readInputFile.initialize();

		environmentVariables.setThreadNumber(readInputFile.getThreadNumber());
		environmentVariables.setOutputDirectoryName(readInputFile.getOutputDirectoryName());
		environmentVariables.setExploreVariables(readInputFile.getExploreVariables());
		environmentVariables.setReOptimizeVariables(readInputFile.getReOptimizeVariables());

		trialBasisCompositions = readInputFile.getTrialBasisCompositions();
	}

	// Read input file
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make directories for output files

	const std::string outputResultParentDirectoryName = environmentVariables.getOutputDirectoryName() + "Result/";
	const std::string outputDensestPackingsParentDirectoryName = environmentVariables.getOutputDirectoryName() + "DensestPackings/";
	makeOutputDirectories(environmentVariables.getOutputDirectoryName(), trialBasisCompositions);

	// make directories for output files
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// set some parameters of CalculationAnalysis

	CalculationAnalysis::setPackingFractionsErrorMesh(environmentVariables.getExploreVariables().getPackingFractionsErrorMesh());
	CalculationAnalysis::setInitialPackingOptimizePackingFractionsMesh(environmentVariables.getExploreVariables().getInitialPackingOptimizePackingFractionsMesh());
	CalculationAnalysis::setPackingOptimizePackingFractionsMesh(environmentVariables.getExploreVariables().getPackingOptimizePackingFractionsMesh());

	// set some parameters of CalculationAnalysis
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// search in the densest packings for each composition

	for (const auto& basisComposition : trialBasisCompositions)
	{
		std::string resultDirectory = outputResultParentDirectoryName;
		std::string densestPackingsDirectory = outputDensestPackingsParentDirectoryName;
		{
			std::string resultDirectoryName;
			std::string densestPackingsDirectoryName;

			for (const auto& radius : basisComposition.getRadii())
			{
				resultDirectoryName += std::to_string(radius) + "-";
				densestPackingsDirectoryName += std::to_string(radius) + "-";
			}

			resultDirectoryName.pop_back();
			densestPackingsDirectoryName.pop_back();
			resultDirectoryName += '/';
			densestPackingsDirectoryName += '/';

			resultDirectory += resultDirectoryName;
			densestPackingsDirectory += densestPackingsDirectoryName;
		}

		{
			std::string resultDirectoryName;
			std::string densestPackingsDirectoryName;

			for (const auto& number : basisComposition.getComposition())
			{
				resultDirectoryName += std::to_string(number.second) + "-";
				densestPackingsDirectoryName += std::to_string(number.second) + "-";
			}

			resultDirectoryName.pop_back();
			densestPackingsDirectoryName.pop_back();
			resultDirectoryName += '/';
			densestPackingsDirectoryName += '/';

			resultDirectory += resultDirectoryName;
			densestPackingsDirectory += densestPackingsDirectoryName;
		}

		IsDirectory isDirectory;
		{
			if (!(isDirectory(resultDirectory)))
			{
				std::string message;
				message += "could not find the ";
				message += resultDirectory;

				throw std::invalid_argument(message);
			}

			if (!(isDirectory(densestPackingsDirectory)))
			{
				std::string message;
				message += "could not find the ";
				message += densestPackingsDirectory;

				throw std::invalid_argument(message);
			}
		}

		TaskManager taskManager(basisComposition, resultDirectory, densestPackingsDirectory);
		taskManager.execute();
	}

	// search in the densest packings for each composition
// ********************************************************************************************************************************************************************

	return 0;
}

void SAMLAIdsp::makeOutputDirectories(const std::string& outputDirectoryName, const std::vector<BasisComposition>& basisCompositions)
{
	const std::string outputResultParentDirectoryName = outputDirectoryName + "Result/";
	const std::string outputDensestPackingsParentDirectoryName = outputDirectoryName + "DensestPackings/";

	{
		MakeDirectory makeDirectory(outputDirectoryName);
		makeDirectory("Result/");
		makeDirectory("DensestPackings/");
	}

	for (const auto& basisComposition : basisCompositions)
	{
		std::string resultDirectory = outputResultParentDirectoryName;
		std::string densestPackingsDirectory = outputDensestPackingsParentDirectoryName;
		{
			std::string resultDirectoryName;
			std::string densestPackingsDirectoryName;

			for (const auto& radius : basisComposition.getRadii())
			{
				resultDirectoryName += std::to_string(radius) + "-";
				densestPackingsDirectoryName += std::to_string(radius) + "-";
			}

			resultDirectoryName.pop_back();
			densestPackingsDirectoryName.pop_back();
			resultDirectoryName += '/';
			densestPackingsDirectoryName += '/';

			MakeDirectory makeResultDirectory(resultDirectory);
			makeResultDirectory(resultDirectoryName);

			MakeDirectory makeDensestPackingsDirectory(densestPackingsDirectory);
			makeDensestPackingsDirectory(densestPackingsDirectoryName);

			resultDirectory += resultDirectoryName;
			densestPackingsDirectory += densestPackingsDirectoryName;
		}

		{
			std::string resultDirectoryName;
			std::string densestPackingsDirectoryName;

			for (const auto& number : basisComposition.getComposition())
			{
				resultDirectoryName += std::to_string(number.second) + "-";
				densestPackingsDirectoryName += std::to_string(number.second) + "-";
			}

			resultDirectoryName.pop_back();
			densestPackingsDirectoryName.pop_back();
			resultDirectoryName += '/';
			densestPackingsDirectoryName += '/';

			MakeDirectory makeResultDirectory(resultDirectory);
			makeResultDirectory(resultDirectoryName);

			MakeDirectory makeDensestPackingsDirectory(densestPackingsDirectory);
			makeDensestPackingsDirectory(densestPackingsDirectoryName);

			resultDirectory += resultDirectoryName;
			densestPackingsDirectory += densestPackingsDirectoryName;
		}

		MakeDirectory makeResultDirectory(resultDirectory);
		makeResultDirectory("DatFiles/");
		makeResultDirectory("CifFiles/");

		MakeDirectory makeDensestPackingDirectory(densestPackingsDirectory);
		makeDensestPackingDirectory("DatFiles/");
		makeDensestPackingDirectory("CifFiles/");
	}
}
