#include "SharedReporter.h"

#include <fstream>

using namespace rk::dnspack;


std::mutex SharedReporter::m_reportMutex;
std::map<std::string, int> SharedReporter::m_sharedNumbers;


SharedReporter::SharedReporter(const std::string& fileName)
	: m_fileName{ fileName }
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	std::map<std::string, int>::iterator sharedNumbersIter = m_sharedNumbers.find(m_fileName);

	if (m_sharedNumbers.end() != sharedNumbersIter)
	{
		sharedNumbersIter->second += 1;
	}

	else
	{
		m_sharedNumbers.emplace(m_fileName, 1);

		std::ofstream outputFileStream;
		outputFileStream.open(m_fileName, std::ios_base::out);

		if (!outputFileStream)
		{
			std::string message = "couldn't open the file of ";
			message += m_fileName;
			message += ".";

			throw std::invalid_argument(message);
		}

		outputFileStream.close();
	}
}

SharedReporter::~SharedReporter()
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	if (!(m_fileName.empty()))
	{
		m_sharedNumbers.at(m_fileName) -= 1;

		if (0 == m_sharedNumbers.at(m_fileName))
		{
			m_sharedNumbers.erase(m_fileName);
		}
	}
}

SharedReporter::SharedReporter(const SharedReporter& other)
	: m_fileName{ other.m_fileName }
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	if (!(m_fileName.empty()))
	{
		m_sharedNumbers.at(m_fileName) += 1;
	}
}

SharedReporter::SharedReporter(SharedReporter&& other) noexcept
	: m_fileName{ other.m_fileName }
{
	other.m_fileName.clear();
}

SharedReporter& SharedReporter::operator=(const SharedReporter& other)
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	if (!(m_fileName.empty()))
	{
		m_sharedNumbers.at(m_fileName) -= 1;

		if (0 == m_sharedNumbers.at(m_fileName))
		{
			m_sharedNumbers.erase(m_fileName);
		}
	}

	m_fileName = other.m_fileName;


	if (!(m_fileName.empty()))
	{
		m_sharedNumbers.at(m_fileName) += 1;
	}

	return *this;
}

SharedReporter& SharedReporter::operator=(SharedReporter&& other) noexcept
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	if (!(m_fileName.empty()))
	{
		m_sharedNumbers.at(m_fileName) -= 1;

		if (0 == m_sharedNumbers.at(m_fileName))
		{
			m_sharedNumbers.erase(m_fileName);
		}
	}

	m_fileName = other.m_fileName;
	other.m_fileName.clear();

	return *this;
}



#ifdef _MSC_VER

#include <iostream>

void SharedReporter::report(const std::string& message) const
{
	std::cout << message;

	std::lock_guard<std::mutex> lock(m_reportMutex);

	std::ofstream outputFileStream;
	outputFileStream.open(m_fileName, std::ios_base::app);

	if (!outputFileStream)
	{
		std::string message = "couldn't open the file of ";
		message += m_fileName;
		message += ".";

		throw std::invalid_argument(message);
	}

	outputFileStream << message;
	outputFileStream.close();
}


#else

void SharedReporter::report(const std::string& message) const
{
	std::lock_guard<std::mutex> lock(m_reportMutex);

	std::ofstream outputFileStream;
	outputFileStream.open(m_fileName, std::ios_base::app);

	if (!outputFileStream)
	{
		std::string message = "couldn't open the file of ";
		message += m_fileName;
		message += ".";

		throw std::invalid_argument(message);
	}

	outputFileStream << message;
	outputFileStream.close();
}


#endif // _MSC_VER
