#include "TaskManager.h"

#include <string>
#include <sstream>
#include <vector>
#include <thread>
#include <mutex>
#include <random>

#include "PackingStructure.h"
#include "UniqueReporter.h"
#include "SharedReporter.h"
#include "SAMLAIReadme.h"

#include "Restart.h"

#include "Clock.h"
#include "PackingFactory.h"
#include "CalculationAnalysis.h"
#include "ExploreSchedule.h"

#include "ReOptimizeFactory.h"
#include "ReOptimizeSchedule.h"


using namespace rk::dnspack;


TaskManager::TaskManager(const BasisComposition& basisComposition, const std::string& resultDirectoryName, const std::string& densestPackingsDirectoryName)
	: m_basisComposition(basisComposition)
	, m_resultDirectoryName(resultDirectoryName)
	, m_densestPackingsDirectoryName(densestPackingsDirectoryName)
	, m_environmentVariables()
{
}

TaskManager::~TaskManager()
{
}

int TaskManager::execute()
{
	Clock clock;

	ExploreVariables exploreVariables = m_environmentVariables.getExploreVariables();
	ReOptimizeVariables reOptimizeVariables = m_environmentVariables.getReOptimizeVariables();

// ********************************************************************************************************************************************************************
	// search in the densest packings

	std::string exploreTimeKey = "exploreTimeKey";
	clock.stopwatchBegin(exploreTimeKey);

	SharedReporter reportExploreSituation(m_resultDirectoryName + "currentSituation.txt");
	{
		std::stringstream reportMessage;

		reportMessage << "*********************************************************************************************************\n\n";
		reportMessage << "This explore of the densest sphere packings was performed by SAMLAIdsp " + SAMLAIVersion + "\n";
		reportMessage << Clock::getNowTime() << "\n\n";
		reportMessage << "*********************************************************************************************************\n\n\n";

		reportMessage << "current time   :  " << clock.getNowTime() << "\n";
		reportMessage << "Thread Number  =  " << m_environmentVariables.getThreadNumber() << "\n\n";
		reportExploreSituation.report(reportMessage.str());
	}

	CalculationAnalysis calculationAnalysis;
	calculationAnalysis.clear();

	ExploreSchedule exploreSchedule(m_resultDirectoryName);
	ExploreSchedule::resetSchedule();
	exploreSchedule.initialDemandReceive();
	exploreSchedule.demandReceive();

	{
		std::string resultFileName = m_resultDirectoryName + "exploreResult.txt";

		double maxPackingFraction = 0.0;
		Restart restart(resultFileName);

		if (restart.readResultFile(maxPackingFraction))
		{
			ExploreSchedule::setMaxPackingFraction(maxPackingFraction);
		}
	}

	std::vector<std::thread> produceTasks;
	produceTasks.reserve(static_cast<std::size_t>(m_environmentVariables.getThreadNumber()));

	std::vector<PackingFactory> packingFactories;
	{
		for (int index = 0; index < m_environmentVariables.getThreadNumber(); ++index)
		{
			packingFactories.push_back(PackingFactory(m_basisComposition, exploreSchedule));
		}
	}

	for (int index = 0; index < m_environmentVariables.getThreadNumber(); ++index)
	{
		produceTasks.push_back(std::thread{ packingFactories[index] });
	}

	for (auto& produceTask : produceTasks)
	{
		produceTask.join();
	}

	packingFactories.clear();

	std::string exploreElapsedTime = clock.getElapsedTime(exploreTimeKey);
	{
		std::stringstream reportMessage;
		reportMessage << "elapsed time for search in densest packings   :  " << exploreElapsedTime << "\n";
		reportExploreSituation.report(reportMessage.str());
	}

	clock.stopwatchFinish(exploreTimeKey);

	// search in the densest packings
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make result file of the exploration

	UniqueReporter makeResultFile(m_resultDirectoryName + "exploreResult.txt");

	{
		std::vector<std::string> sphereNames(6);
		{
			sphereNames[0] = "H";
			sphereNames[1] = "Na";
			sphereNames[2] = "Sc";
			sphereNames[3] = "Y";
			sphereNames[4] = "La";
			sphereNames[5] = "Hf";
		}

		std::vector<double> radii;
		{
			for (const auto& radius : m_basisComposition.getRadii())
			{
				radii.push_back(radius);
			}
		}

		if (6 < radii.size())
		{
			throw std::runtime_error("The kind of sphere is larger than 6.");
		}

		std::stringstream message;

		message << "*********************************************************************************************************\n\n";
		message << "This explore of the densest sphere packings was performed by SAMLAIdsp " + SAMLAIVersion + "\n";
		message << Clock::getNowTime() << "\n\n";
		message << "*********************************************************************************************************\n\n";

		message << "curent time  :  " << Clock::getNowTime() << "\n\n\n";

		message << "#\n" << "# Spheres Information : The name and the radius of each sphere\n" << "#\n\n\n";
		message << "<Definition.of.Sphere.Species\n";

		for (std::size_t label = 0; label < m_basisComposition.getSphereKindNumber(); ++label)
		{
			message << " " << sphereNames[label] << "    " << radii[label] << "\n";
		}
		message << "Definition.of.Sphere.Species>\n\n";

		message << "Atoms.Number   " << m_basisComposition.getSpheresNumber() << "\n";
		message << "<Number.Of.Spheres.Per.UnitCell\n";
		{
			std::size_t label = 0;

			for (const auto& number : m_basisComposition.getComposition())
			{
				message << " " << sphereNames[label] << "    " << number.second << "\n";
				++label;
			}
		}
		message << "Number.Of.Spheres.Per.UnitCell>\n";

		message << "\n\n#\n" << "# Optimize Paramaters for search in the densest packings\n" << "#\n\n\n";
		message << "Annealing.Maximum.Displacement.Upper                               " << exploreVariables.getAnnealingMaxDisplacementUpper() << "\n";
		message << "Annealing.Maximum.Displacement.Lower                               " << exploreVariables.getAnnealingMaxDisplacementLower() << "\n";
		message << "Annealing.Iteration.Number                                         " << exploreVariables.getAnnealingIterationNumber() << "\n";
		message << "Annealing.Interval.To.Update.The.Neighboringlist                   " << exploreVariables.getAnnealingNeighborUpdateInterval() << "\n\n";

		message << "InitialPackingOptimize.Decrease.Factor                             " << exploreVariables.getInitialPackingOptimizeDecreaseFactor() << "\n";
		message << "InitialPackingOptimize.Iteration.Number                            " << exploreVariables.getInitialPackingOptimizeIterationNumber() << "\n";
		message << "InitialPackingOptimize.Interval.To.Update.The.Neighboringlist      " << exploreVariables.getInitialPackingOptimizeNeighborUpdateInterval() << "\n\n";

		message << "PackingOptimize.Maximum.Displacement.Upper                         " << exploreVariables.getPackingOptimizeMaxDisplacementUpper() << "\n";
		message << "PackingOptimize.Maximum.Displacement.Lower                         " << exploreVariables.getPackingOptimizeMaxDisplacementLower() << "\n";
		message << "PackingOptimize.Decrease.Factor                                    " << exploreVariables.getPackingOptimizeDecreaseFactor() << "\n";
		message << "PackingOptimize.Iteration.Number                                   " << exploreVariables.getPackingOptimizeIterationNumber() << "\n";
		message << "PackingOptimize.Interval.To.Update.The.Neighboringlist             " << exploreVariables.getPackingOptimizeNeighborUpdateInterval() << "\n\n";

		message << "\n\n#\n" << "# Conditions for search in the densest packings and for the output\n" << "#\n\n\n";

		message << "FullOptimize.Structure.Number.In.Beginning                         " << exploreSchedule.getFullOptimizeStructureNumber() << "\n";
		message << "Maximum.Trial.Number.For.Each.Composition                          " << exploreSchedule.getMaxTrialNumber() << "\n";
		message << "Threshold.Trial.Number.For.Judging.Convergence                     " << exploreSchedule.getThresholdTrialNumber() << "\n";

		message << "Mesh.For.Packing.Fractions.Error                                   " << exploreVariables.getPackingFractionsErrorMesh() << "\n";
		message << "Mesh.For.InitialPackingOptimize.Packing.Fractions                  " << exploreVariables.getInitialPackingOptimizePackingFractionsMesh() << "\n";
		message << "Mesh.For.PackingOptimize.Packing.Fractions                         " << exploreVariables.getPackingOptimizePackingFractionsMesh() << "\n";

		message << "Output.Minimum.Packing.Fraction                                    " << exploreSchedule.getOutputMinimumPackingFraction() << "\n";
		message << "Output.Minimum.Overlap                                             " << exploreSchedule.getOutputMinimumZ() << "\n\n";

		message << "\n#\n" << "# The result of the exploration\n" << "#\n\n\n";

		message << "Elapsed.Time.of.The.Exploration                                    " << exploreElapsedTime << "\n";
		message << "Number.Of.Initial.Generated.Structures                             " << exploreSchedule.getInitialDemandNumber() << "\n";
		message << "Number.Of.Generated.Structures                                     " << exploreSchedule.getDemandNumber() << "\n";

		message << "Low.Packing.Fraction.Reject.Number.After.SmallAnnealing            " << calculationAnalysis.getLowPackingFractionRejectNumber() << "\n";
		message << "Final.Threshold.SmallAnnealing.Packing.Fraction                    " << calculationAnalysis.getThresholdPackingFraction() << "\n";
		message << "Maximum.Packing.Fraction                                           " << exploreSchedule.getMaxPackingFraction() << "\n";
		message << "Number.Of.The.Generated.Putative.Densest.Structure                 " << exploreSchedule.getMaxPackingFractionStructureNumber() << "\n\n";

		message << "Annealing.MinusVolume.Reject.Number                                " << calculationAnalysis.getAnnealingMinusVolumeRejectNumber() << "\n";
		message << "InitialPackingOptimize.MinusVolume.Reject.Number                   " << calculationAnalysis.getInitialPackingOptimizeMinusVolumeRejectNumber() << "\n";
		message << "SpherePackingOptimize.MinusVolume.Reject.Number                    " << calculationAnalysis.getPackingOptimizeMinusVolumeRejectNumber() << "\n";
		message << "SpherePackingOptimize.Zvalue.Reject.Number                         " << calculationAnalysis.getPackingOptimizeZvalueRejectNumber() << "\n\n\n";

		makeResultFile.report(message.str());
	}

	// make result file of the exploration
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make file for the error distribution of packing fractions

	UniqueReporter packingFractionsErrorDistributionFile(m_resultDirectoryName + "errorDistribution.txt");

	{
		std::stringstream message;

		for (const auto& distribution : calculationAnalysis.getPackingFractionsErrorDistribution())
		{
			message << distribution.first << "   " << distribution.second << "\n";
		}

		packingFractionsErrorDistributionFile.report(message.str());
	}

	// make file for the error distribution of packing fractions
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make file for the distribution of packing fractions in SmallAnnealing

	UniqueReporter smallAnnealingPackingFractionsDistributionFile(m_resultDirectoryName + "smallAnnealingDistribution.txt");

	{
		std::stringstream message;

		for (const auto& distribution : calculationAnalysis.getInitialPackingOptimizePackingFractionsDistribution())
		{
			message << distribution.first << "   " << distribution.second << "\n";
		}

		smallAnnealingPackingFractionsDistributionFile.report(message.str());
	}

	// make file for the distribution of packing fractions in SmallAnnealing
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make file for the distribution of packing fractions in PackingOptimize

	UniqueReporter packingOptimizePackingFractionsDistributionFile(m_resultDirectoryName + "packingOptimizeDistribution.txt");

	{
		std::stringstream message;

		for (const auto& distribution : calculationAnalysis.getPackingOptimizePackingFractionsDistribution())
		{
			message << distribution.first << "   " << distribution.second << "\n";
		}

		packingOptimizePackingFractionsDistributionFile.report(message.str());
	}

	// make file for the distribution of packing fractions in PackingOptimize
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// re-optimization

	std::string reOptimizationTimeKey = "reOptimizationTimeKey";
	clock.stopwatchBegin(reOptimizationTimeKey);

	SharedReporter reportReOptimizeSituation(m_densestPackingsDirectoryName + "currentSituation.txt");
	{
		std::stringstream reportMessage;

		reportMessage << "*********************************************************************************************************\n\n";
		reportMessage << "This explore of the densest sphere packings was performed by SAMLAIdsp " + SAMLAIVersion + "\n";
		reportMessage << Clock::getNowTime() << "\n\n";
		reportMessage << "*********************************************************************************************************\n\n\n";

		reportMessage << "current time   :  " << clock.getNowTime() << "\n";
		reportMessage << "Thread Number  =  " << m_environmentVariables.getThreadNumber() << "\n\n";
		reportReOptimizeSituation.report(reportMessage.str());
	}

	std::vector<PackingStructure> seedStructures;
	{
		for (const auto& pairs : exploreSchedule.getDensePackingStructures())
		{
			seedStructures.push_back(pairs.second);
		}

		if (reOptimizeVariables.getReOptimizeStructureNumber() < seedStructures.size())
		{
			throw std::logic_error("seedStructures contains more structures than m_reOptimizeConditions.getReOptimizeStructureNumber().");
		}

		else
		{
			std::vector<PackingStructure> densePackingStructures = seedStructures;

			std::random_device seed_gen;
			std::default_random_engine engine(seed_gen());
			std::uniform_int_distribution<int> labelDist(0, (static_cast<int>(densePackingStructures.size()) - 1));

			for (std::size_t size = seedStructures.size(); size < reOptimizeVariables.getReOptimizeStructureNumber(); ++size)
			{
				int label = labelDist(engine);
				seedStructures.push_back(densePackingStructures[label]);
			}
		}
	}

	double reOptimizedMaxPackingFraction = 0.0;
	{
		double maxPackingFraction = 0.0;

		{
			std::string resultFileName = m_densestPackingsDirectoryName + "reOptimizationResult.txt";

			Restart restart(resultFileName);

			if (restart.readResultFile(maxPackingFraction))
			{
				ExploreSchedule::setMaxPackingFraction(maxPackingFraction);
			}
		}

		for (const auto& packingStructure : seedStructures)
		{
			ReOptimizeSchedule reOptimizeSchedule(packingStructure, m_densestPackingsDirectoryName);
			ReOptimizeSchedule::resetSchedule();
			reOptimizeSchedule.demandReceive();
			reOptimizeSchedule.setMaxPackingFraction(maxPackingFraction);

			std::vector<std::thread> reOptimizeTasks;
			reOptimizeTasks.reserve(static_cast<std::size_t>(m_environmentVariables.getThreadNumber()));

			std::vector<ReOptimizeFactory> reOptimizeFactories;
			{
				for (int index = 0; index < m_environmentVariables.getThreadNumber(); ++index)
				{
					reOptimizeFactories.push_back(ReOptimizeFactory(reOptimizeSchedule));
				}
			}

			for (int index = 0; index < m_environmentVariables.getThreadNumber(); ++index)
			{
				reOptimizeTasks.push_back(std::thread{ reOptimizeFactories[index] });
			}

			for (auto& reOptimizeTask : reOptimizeTasks)
			{
				reOptimizeTask.join();
			}

			maxPackingFraction = reOptimizeSchedule.getMaxPackingFraction();
		}

		reOptimizedMaxPackingFraction = maxPackingFraction;
	}

	std::string reOptimizationElapsedTime = clock.getElapsedTime(reOptimizationTimeKey);
	{
		std::stringstream reportMessage;
		reportMessage << "elapsed time for search in densest packings   :   " << clock.getElapsedTime(reOptimizationTimeKey) << "\n";
		reportReOptimizeSituation.report(reportMessage.str());
	}

	clock.stopwatchFinish(reOptimizationTimeKey);

	// re-optimization
// ********************************************************************************************************************************************************************

// ********************************************************************************************************************************************************************
	// make result file of the re-optimization

	UniqueReporter makeReOptimizationFile(m_densestPackingsDirectoryName + "reOptimizationResult.txt");

	{
		std::vector<std::string> sphereNames(6);
		{
			sphereNames[0] = "H";
			sphereNames[1] = "Na";
			sphereNames[2] = "Sc";
			sphereNames[3] = "Y";
			sphereNames[4] = "La";
			sphereNames[5] = "Hf";
		}

		std::vector<double> radii;
		{
			for (const auto& radius : m_basisComposition.getRadii())
			{
				radii.push_back(radius);
			}
		}

		if (6 < radii.size())
		{
			throw std::runtime_error("The kind of sphere is larger than 6.");
		}

		std::stringstream message;

		message << "*********************************************************************************************************\n\n";
		message << "This explore of the densest sphere packings was performed by SAMLAIdsp " + SAMLAIVersion + "\n";
		message << Clock::getNowTime() << "\n\n";
		message << "*********************************************************************************************************\n\n";

		message << "curent time  :  " << Clock::getNowTime() << "\n\n\n";

		message << "#\n" << "# Spheres Information : The name and the radius of each sphere\n" << "#\n\n\n";
		message << "<Definition.of.Sphere.Species\n";

		for (std::size_t label = 0; label < m_basisComposition.getSphereKindNumber(); ++label)
		{
			message << " " << sphereNames[label] << "    " << radii[label] << "\n";
		}
		message << "Definition.of.Sphere.Species>\n\n";

		message << "Atoms.Number   " << m_basisComposition.getSpheresNumber() << "\n";
		message << "<Number.Of.Spheres.Per.UnitCell\n";
		{
			std::size_t label = 0;

			for (const auto& number : m_basisComposition.getComposition())
			{
				message << " " << sphereNames[label] << "    " << number.second << "\n";
				++label;
			}
		}
		message << "Number.Of.Spheres.Per.UnitCell>\n";

		message << "\n\n#\n" << "# Re-optimize parameters\n" << "#\n\n\n";
		message << "Maximum.Fluctuation                                                    " << reOptimizeVariables.getMaxFluctuation() << "\n\n";

		message << "InitialAnnealing.Maximum.Displacement.Upper                            " << reOptimizeVariables.getInitialAnnealingMaxDisplacementUpper() << "\n";
		message << "InitialAnnealing.Maximum.Displacement.Lower                            " << reOptimizeVariables.getInitialAnnealingMaxDisplacementLower() << "\n";
		message << "InitialAnnealing.Iteration.Number                                      " << reOptimizeVariables.getInitialAnnealingIterationNumber() << "\n";
		message << "InitialAnnealing.Interval.To.Update.The.Neighboringlist                " << reOptimizeVariables.getInitialAnnealingNeighborUpdateInterval() << "\n\n";

		message << "DensestPackingOptimize.Maximum.Displacement.Upper                      " << reOptimizeVariables.getDensestPackingOptimizeMaxDisplacementUpper() << "\n";
		message << "DensestPackingOptimize.Maximum.Displacement.Lower                      " << reOptimizeVariables.getDensestPackingOptimizeMaxDisplacementLower() << "\n";
		message << "DensestPackingOptimize.Decrease.Factor                                 " << reOptimizeVariables.getDensestPackingOptimizeDecreaseFactor() << "\n";
		message << "DensestPackingOptimize.Iteration.Number                                " << reOptimizeVariables.getDensestPackingOptimizeIterationNumber() << "\n";
		message << "DensestPackingOptimize.Interval.To.Update.The.Neighboringlist          " << reOptimizeVariables.getDensestPackingOptimizeNeighborUpdateInterval() << "\n\n";

		message << "\n\n#\n" << "# Conditions for re-optimization and the output\n" << "#\n\n\n";

		message << "Maximum.Packing.Fraction                                               " << reOptimizedMaxPackingFraction << "\n";
		message << "Number.Of.ReOptimize.Structures                                        " << reOptimizeVariables.getReOptimizeStructureNumber() << "\n";
		message << "Number.Of.Re-Optimization.For.Each.Structure                           " << reOptimizeVariables.getReOptimizeRepetitionNumber() << "\n";
		message << "Output.Minimum.Overlap.For.ReOptimization                              " << reOptimizeVariables.getOutputMinimumZ() << "\n\n";

		message << "\n#\n" << "# The result of the exploration\n" << "#\n\n\n";

		message << "Elapsed.Time.of.The.Exploration                                        " << reOptimizationElapsedTime << "\n";

		makeReOptimizationFile.report(message.str());
	}

	// make result file of the exploration
// ********************************************************************************************************************************************************************

	return 0;
}
